# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.optimize import minimize

from .optimization import Optimization
from .derive_downstream_turbines import derive_downstream_turbines


class YawOptimizationWindRose(Optimization):
    """
    YawOptimizationWindRose is a subclass of
    :py:class:`~.tools.optimization.scipy.Optimization` that is used to
    optimize the yaw angles of all turbines in a Floris Farm for multiple sets
    of inflow conditions (combinations of wind speed, wind direction, and
    optionally turbulence intensity) using the scipy optimize package.
    """

    def __init__(
        self,
        fi,
        wd,
        ws,
        ti=None,
        minimum_yaw_angle=0.0,
        maximum_yaw_angle=25.0,
        minimum_ws=3.0,
        maximum_ws=25.0,
        yaw_angles_baseline=None,
        x0=None,
        bnds=None,
        opt_method="SLSQP",
        opt_options=None,
        include_unc=False,
        unc_pmfs=None,
        unc_options=None,
        turbine_weights=None,
        verbose=False,
        calc_init_power=True,
        exclude_downstream_turbines=False,
    ):
        """
        Instantiate YawOptimizationWindRose object with a FlorisInterface
        object and assign parameter values.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            wd (iterable) : The wind directions for which the yaw angles are
                optimized (deg).
            ws (iterable): The wind speeds for which the yaw angles are
                optimized (m/s).
            ti (iterable, optional): An optional list of turbulence intensity
                values for which the yaw angles are optimized. If not
                specified, the current TI value in the Floris object will be
                used for all optimizations. Defaults to None.
            minimum_yaw_angle (float, optional): Minimum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to 0.0.
            maximum_yaw_angle (float, optional): Maximum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to 25.0.
            minimum_ws (float, optional): Minimum wind speed at which
                optimization is performed (m/s). Assumes zero power generated
                below this value. Defaults to 3.
            maximum_ws (float, optional): Maximum wind speed at which
                optimization is performed (m/s). Assumes optimal yaw offsets
                are zero above this wind speed. Defaults to 25.
            yaw_angles_baseline (iterable, optional): The baseline yaw
                angles used to calculate the initial and baseline power
                production in the wind farm and used to normalize the cost
                function. If none are specified, this variable is set equal
                to the current yaw angles in floris. Note that this variable
                need not meet the yaw constraints specified in self.bnds,
                yet a warning is raised if it does to inform the user.
                Defaults to None.
            x0 (iterable, optional): The initial guess for the optimization
                problem. These values must meet the constraints specified
                in self.bnds. Note that, if exclude_downstream_turbines=True,
                the initial guess for any downstream turbines are ignored
                since they are not part of the optimization. Instead, the yaw
                angles for those turbines are 0.0 if that meets the lower and
                upper bound, or otherwise as close to 0.0 as feasible. If no
                values for x0 are specified, x0 is set to be equal to zeros
                wherever feasible (w.r.t. the bounds), and equal to the
                average of its lower and upper bound for all non-downstream
                turbines otherwise. Defaults to None.
            bnds (iterable, optional): Bounds for the yaw angles, as tuples of
                min, max values for each turbine (deg). One can fix the yaw
                angle of certain turbines to a predefined value by setting that
                turbine's lower bound equal to its upper bound (i.e., an
                equality constraint), as: bnds[ti] = (x, x), where x is the
                fixed yaw angle assigned to the turbine. This works for both
                zero and nonzero yaw angles. Moreover, if 
                exclude_downstream_turbines=True, the yaw angles for all
                downstream turbines will be 0.0 or a feasible value closest to
                0.0. If none are specified, the bounds are set to
                (minimum_yaw_angle, maximum_yaw_angle) for each turbine. Note
                that, if bnds is not none, its values overwrite any value given
                in minimum_yaw_angle and maximum_yaw_angle. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dictionary, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': False, 'iprint': 1, 'ftol': 1e-7,
                'eps': 0.01}. Defaults to None.
            include_unc (bool, optional): Determines whether wind direction or
                yaw uncertainty are included. If True, uncertainty in wind
                direction and/or yaw position is included when determining
                wind farm power. Uncertainty is included by computing the
                mean wind farm power for a distribution of wind direction
                and yaw position deviations from the intended wind direction
                and yaw angles. Defaults to False.
            unc_pmfs (dictionary, optional): A dictionary containing
                probability mass functions describing the distribution of
                wind direction and yaw position deviations when wind direction
                and/or yaw position uncertainty is included in the power
                calculations. Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): The wind direction
                    deviations from the intended wind direction (deg).
                -   **wd_unc_pmf** (*np.array*): The probability
                    of each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): The yaw angle deviations
                    from the intended yaw angles (deg).
                -   **yaw_unc_pmf** (*np.array*): The probability
                    of each yaw angle deviation in **yaw_unc** occuring.

                If none are specified, default PMFs are calculated using
                values provided in **unc_options**. Defaults to None.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): The standard deviation of
                    the wind direction deviations from the original wind
                    direction (deg).
                -   **std_yaw** (*float*): The standard deviation of
                    the yaw angle deviations from the original yaw angles (deg).
                -   **pmf_res** (*float*): The resolution in degrees
                    of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): The cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                If none are specified, default values of
                {'std_wd': 4.95, 'std_yaw': 1.75, 'pmf_res': 1.0,
                'pdf_cutoff': 0.995} are used. Defaults to None.
            turbine_weights (iterable, optional): weighing terms that allow
                the user to emphasize power gains at particular turbines or
                completely ignore power gains from other turbines. The array
                of turbine powers from floris is multiplied with this array
                in the calculation of the objective function. If None, this
                is an array with all values 1.0 and length equal to the
                number of turbines. Defaults to None.
            calc_init_power (bool, optional): If True, calculates initial
                wind farm power for each set of wind conditions. Defaults to
                True.
            exclude_downstream_turbines (bool, optional): If True,
                automatically finds and excludes turbines that are most
                downstream from the optimization problem. This significantly
                reduces computation time at no loss in performance. The yaw
                angles of these downstream turbines are fixed to 0.0 deg if
                the yaw bounds specified in self.bnds allow that, or otherwise
                are fixed to the lower or upper yaw bound, whichever is closer
                to 0.0. Defaults to False.
        """
        super().__init__(fi)

        if opt_options is None:
            self.opt_options = {
                "maxiter": 100,
                "disp": False,
                "iprint": 1,
                "ftol": 1e-7,
                "eps": 0.01,
            }

        self.unc_pmfs = unc_pmfs

        if unc_options is None:
            self.unc_options = {
                "std_wd": 4.95,
                "std_yaw": 1.75,
                "pmf_res": 1.0,
                "pdf_cutoff": 0.995,
            }

        self.ti = ti

        self.reinitialize_opt_wind_rose(
            wd=wd,
            ws=ws,
            ti=ti,
            minimum_yaw_angle=minimum_yaw_angle,
            maximum_yaw_angle=maximum_yaw_angle,
            minimum_ws=minimum_ws,
            maximum_ws=maximum_ws,
            yaw_angles_baseline=yaw_angles_baseline,
            x0=x0,
            bnds=bnds,
            opt_method=opt_method,
            opt_options=opt_options,
            include_unc=include_unc,
            unc_pmfs=unc_pmfs,
            unc_options=unc_options,
            turbine_weights=turbine_weights,
            calc_init_power=calc_init_power,
            exclude_downstream_turbines=exclude_downstream_turbines,
        )

        self.verbose = verbose

    # Private methods

    def _get_initial_farm_power(self):
        self.initial_farm_powers = []

        for i in range(len(self.wd)):
            if (self.ws[i] >= self.minimum_ws) & (self.ws[i] <= self.maximum_ws):
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )

                # initial power
                self.fi.calculate_wake(yaw_angles=self.yaw_angles_baseline)
                power_init = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            elif self.ws[i] >= self.maximum_ws:
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )
                self.fi.calculate_wake(yaw_angles=self.yaw_angles_baseline)
                power_init = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            else:
                power_init = self.nturbs * [0.0]

            self.initial_farm_powers.append(np.dot(self.turbine_weights, power_init))

    def _get_power_for_yaw_angle_opt(self, yaw_angles_subset_norm):
        """
        Assign yaw angles to turbines, calculate wake, report power

        Args:
            yaw_angles_subset_norm (np.array): Yaw to apply to subset
            of controlled turbines, normalized.

        Returns:
            power (float): Wind plant power. #TODO negative? in kW?
        """
        yaw_angles_subset = self._unnorm(
            np.array(yaw_angles_subset_norm),
            self.minimum_yaw_angle,
            self.maximum_yaw_angle,
        )

        # Create a full yaw angle array
        yaw_angles = np.array(self.yaw_angles_template, copy=True)
        yaw_angles[self.turbs_to_opt] = yaw_angles_subset

        self.fi.calculate_wake(yaw_angles=yaw_angles)
        turbine_powers = self.fi.get_turbine_power(
            include_unc=self.include_unc,
            unc_pmfs=self.unc_pmfs,
            unc_options=self.unc_options,
        )

        return (
            -1.0
            * np.dot(self.turbine_weights, turbine_powers)
            / self.initial_farm_power
        )

    def _set_opt_bounds(self, minimum_yaw_angle, maximum_yaw_angle):
        """
        Sets minimum and maximum yaw angle bounds for optimization.
        """

        self.bnds = [(minimum_yaw_angle, maximum_yaw_angle) for _ in range(self.nturbs)]

    def _optimize(self):
        """
        Find optimum setting of turbine yaw angles for power production
        given fixed atmospheric conditions (wind speed, direction, etc.).

        Returns:
            opt_yaw_angles (np.array): Optimal yaw angles of each turbine.
        """
        opt_yaw_angles = np.array(self.yaw_angles_template, copy=True)
        wind_map = self.fi.floris.farm.wind_map
        self._reduce_control_variables()

        if len(self.turbs_to_opt) > 0:
            self.residual_plant = minimize(
                self._get_power_for_yaw_angle_opt,
                self.x0_norm,
                method=self.opt_method,
                bounds=self.bnds_norm,
                options=self.opt_options,
            )
            opt_yaw_angles_subset = self._unnorm(
                self.residual_plant.x, self.minimum_yaw_angle, self.maximum_yaw_angle
            )
            opt_yaw_angles[self.turbs_to_opt] = opt_yaw_angles_subset

        self.fi.reinitialize_flow_field(
            wind_speed=wind_map.input_speed,
            wind_direction=wind_map.input_direction,
            turbulence_intensity=wind_map.input_ti,
        )
        return opt_yaw_angles

    def _reduce_control_variables(self):
        """This function reduces the control problem by eliminating turbines
        of which the yaw angles need not be optimized, either because of a
        user-specified set of bounds (where bounds[i][0] == bounds[i][1]),
        or alternatively turbines that are far downstream in the wind farm
        and of which the wake does not impinge other turbines, if the
        boolean exclude_downstream_turbines == True. The normalized initial
        conditions and bounds are then calculated for the subset of turbines,
        to be used in the optimization.
        """
        if self.bnds is not None:
            self.turbs_to_opt, _ = np.where(np.abs(np.diff(self.bnds)) >= 0.001)
        else:
            self.turbs_to_opt = np.array(range(self.nturbs), dtype=int)

        if self.exclude_downstream_turbines:
            # Remove turbines from turbs_to_opt that are downstream
            downstream_turbines = derive_downstream_turbines(
                fi=self.fi, wind_direction=self.fi.floris.farm.wind_direction[0]
            )
            downstream_turbines = np.array(downstream_turbines, dtype=int)
            self.turbs_to_opt = [
                i for i in self.turbs_to_opt if i not in downstream_turbines
            ]

        # Set up a template yaw angles array with default solutions. The default
        # solutions are either 0.0 or the allowable yaw angle closest to 0.0 deg.
        # This solution addresses both downstream turbines, minimizing their abs.
        # yaw offset, and additionally fixing equality-constrained turbines to
        # their appropriate yaw angle.
        yaw_angles_template = np.zeros(self.nturbs, dtype=float)
        for ti in range(self.nturbs):
            if (self.bnds[ti][0] > 0.0) | (self.bnds[ti][1] < 0.0):
                yaw_angles_template[ti] = self.bnds[ti][
                    np.argmin(np.abs(self.bnds[ti]))
                ]
        self.yaw_angles_template = yaw_angles_template

        # Derive normalized initial condition and bounds
        x0_subset = [self.x0[i] for i in self.turbs_to_opt]
        self.x0_norm = self._norm(
            np.array(x0_subset), self.minimum_yaw_angle, self.maximum_yaw_angle
        )
        self.bnds_norm = [
            (
                self._norm(
                    self.bnds[i][0], self.minimum_yaw_angle, self.maximum_yaw_angle
                ),
                self._norm(
                    self.bnds[i][1], self.minimum_yaw_angle, self.maximum_yaw_angle
                ),
            )
            for i in self.turbs_to_opt
        ]

    # Public methods

    def reinitialize_opt_wind_rose(
        self,
        wd=None,
        ws=None,
        ti=None,
        minimum_yaw_angle=None,
        maximum_yaw_angle=None,
        minimum_ws=None,
        maximum_ws=None,
        yaw_angles_baseline=None,
        x0=None,
        bnds=None,
        opt_method=None,
        opt_options=None,
        include_unc=None,
        unc_pmfs=None,
        unc_options=None,
        turbine_weights=None,
        calc_init_power=True,
        exclude_downstream_turbines=None,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            wd (iterable, optional) : The wind directions for which the yaw
                angles are optimized (deg). Defaults to None.
            ws (iterable, optional): The wind speeds for which the yaw angles
                are optimized (m/s). Defaults to None.
            ti (iterable, optional): An optional list of turbulence intensity
                values for which the yaw angles are optimized. Defaults to None.
            minimum_yaw_angle (float, optional): Minimum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to None.
            maximum_yaw_angle (float, optional): Maximum constraint on yaw
                angle (deg). This value will be ignored if bnds is also
                specified. Defaults to None.
            minimum_ws (float, optional): Minimum wind speed at which
                optimization is performed (m/s). Assumes zero power generated
                below this value. Defaults to None.
            maximum_ws (float, optional): Maximum wind speed at which
                optimization is performed (m/s). Assumes optimal yaw offsets
                are zero above this wind speed. Defaults to None.
            yaw_angles_baseline (iterable, optional): The baseline yaw
                angles used to calculate the initial and baseline power
                production in the wind farm and used to normalize the cost
                function. If none are specified, this variable is set equal
                to the current yaw angles in floris. Note that this variable
                need not meet the yaw constraints specified in self.bnds,
                yet a warning is raised if it does to inform the user.
                Defaults to None.
            x0 (iterable, optional): The initial guess for the optimization
                problem. These values must meet the constraints specified
                in self.bnds. Note that, if exclude_downstream_turbines=True,
                the initial guess for any downstream turbines are ignored
                since they are not part of the optimization. Instead, the yaw
                angles for those turbines are 0.0 if that meets the lower and
                upper bound, or otherwise as close to 0.0 as feasible. If no
                values for x0 are specified, x0 is set to be equal to zeros
                wherever feasible (w.r.t. the bounds), and equal to the
                average of its lower and upper bound for all non-downstream
                turbines otherwise. Defaults to None.
            bnds (iterable, optional): Bounds for the yaw angles, as tuples of
                min, max values for each turbine (deg). One can fix the yaw
                angle of certain turbines to a predefined value by setting that
                turbine's lower bound equal to its upper bound (i.e., an
                equality constraint), as: bnds[ti] = (x, x), where x is the
                fixed yaw angle assigned to the turbine. This works for both
                zero and nonzero yaw angles. Moreover, if 
                exclude_downstream_turbines=True, the yaw angles for all
                downstream turbines will be 0.0 or a feasible value closest to
                0.0. If none are specified, the bounds are set to
                (minimum_yaw_angle, maximum_yaw_angle) for each turbine. Note
                that, if bnds is not none, its values overwrite any value given
                in minimum_yaw_angle and maximum_yaw_angle. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to None.
            opt_options (dictionary, optional): Optimization options used by
                scipy.optimize.minize. Defaults to None.
            include_unc (bool, optional): Determines whether wind direction or
                yaw uncertainty are included. If True, uncertainty in wind
                direction and/or yaw position is included when determining
                wind farm power. Uncertainty is included by computing the
                mean wind farm power for a distribution of wind direction
                and yaw position deviations from the intended wind direction
                and yaw angles. Defaults to None.
            unc_pmfs (dictionary, optional): A dictionary containing
                probability mass functions describing the distribution of
                wind direction and yaw position deviations when wind direction
                and/or yaw position uncertainty is included in the power
                calculations. Contains the following key-value pairs:

                -   **wd_unc** (*np.array*): The wind direction
                    deviations from the intended wind direction (deg).
                -   **wd_unc_pmf** (*np.array*): The probability
                    of each wind direction deviation in **wd_unc** occuring.
                -   **yaw_unc** (*np.array*): The yaw angle deviations
                    from the intended yaw angles (deg).
                -   **yaw_unc_pmf** (*np.array*): The probability
                    of each yaw angle deviation in **yaw_unc** occuring.

                If none are specified, default PMFs are calculated using
                values provided in **unc_options**. Defaults to None.
            unc_options (dictionary, optional): A dictionary containing values
                used to create normally-distributed, zero-mean probability mass
                functions describing the distribution of wind direction and yaw
                position deviations when wind direction and/or yaw position
                uncertainty is included. This argument is only used when
                **unc_pmfs** is None and contains the following key-value pairs:

                -   **std_wd** (*float*): The standard deviation of
                    the wind direction deviations from the original wind
                    direction (deg).
                -   **std_yaw** (*float*): The standard deviation of
                    the yaw angle deviations from the original yaw angles (deg).
                -   **pmf_res** (*float*): The resolution in degrees
                    of the wind direction and yaw angle PMFs.
                -   **pdf_cutoff** (*float*): The cumulative
                    distribution function value at which the tails of the
                    PMFs are truncated.

                If none are specified, default values of
                {'std_wd': 4.95, 'std_yaw': 1.75, 'pmf_res': 1.0,
                'pdf_cutoff': 0.995} are used. Defaults to None.
            turbine_weights (iterable, optional): weighing terms that allow
                the user to emphasize power gains at particular turbines or
                completely ignore power gains from other turbines. The array
                of turbine powers from floris is multiplied with this array
                in the calculation of the objective function. If None, this
                is an array with all values 1.0 and length equal to the
                number of turbines. Defaults to None.
            calc_init_power (bool, optional): If True, calculates initial
                wind farm power for each set of wind conditions. Defaults to
                None.
            exclude_downstream_turbines (bool, optional): If True,
                automatically finds and excludes turbines that are most
                downstream from the optimization problem. This significantly
                reduces computation time at no loss in performance. The yaw
                angles of these downstream turbines are fixed to 0.0 deg if
                the yaw bounds specified in self.bnds allow that, or otherwise
                are fixed to the lower or upper yaw bound, whichever is closer
                to 0.0. Defaults to None.
        """

        if wd is not None:
            self.wd = wd
        if ws is not None:
            self.ws = ws
        if ti is not None:
            self.ti = ti
        if minimum_ws is not None:
            self.minimum_ws = minimum_ws
        if maximum_ws is not None:
            self.maximum_ws = maximum_ws
        if minimum_yaw_angle is not None:
            self.minimum_yaw_angle = minimum_yaw_angle
        if maximum_yaw_angle is not None:
            self.maximum_yaw_angle = maximum_yaw_angle
        if opt_method is not None:
            self.opt_method = opt_method
        if opt_options is not None:
            self.opt_options = opt_options
        if yaw_angles_baseline is not None:
            self.yaw_angles_baseline = yaw_angles_baseline
        else:
            self.yaw_angles_baseline = [
                turbine.yaw_angle
                for turbine in self.fi.floris.farm.turbine_map.turbines
            ]
            if any(np.abs(self.yaw_angles_baseline) > 0.0):
                print(
                    "INFO: Baseline yaw angles were not specified and were derived from the floris object."
                )
                print(
                    "INFO: The inherent yaw angles in the floris object are not all 0.0 degrees."
                )

        self.bnds = bnds
        if bnds is not None:
            self.minimum_yaw_angle = np.min([bnds[i][0] for i in range(self.nturbs)])
            self.maximum_yaw_angle = np.max([bnds[i][1] for i in range(self.nturbs)])
        else:
            self._set_opt_bounds(self.minimum_yaw_angle, self.maximum_yaw_angle)

        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = np.zeros(self.nturbs, dtype=float)
            for ti in range(self.nturbs):
                if (self.bnds[ti][0] > 0.0) | (self.bnds[ti][1] < 0.0):
                    self.x0[ti] = np.mean(self.bnds[ti])

        if any(
            np.array(self.yaw_angles_baseline) < np.array([b[0] for b in self.bnds])
        ):
            print("INFO: yaw_angles_baseline exceed lower bound constraints.")
        if any(
            np.array(self.yaw_angles_baseline) > np.array([b[1] for b in self.bnds])
        ):
            print("INFO: yaw_angles_baseline in FLORIS exceed upper bound constraints.")
        if any(np.array(self.x0) < np.array([b[0] for b in self.bnds])):
            raise ValueError("Initial guess x0 exceeds lower bound constraints.")
        if any(np.array(self.x0) > np.array([b[1] for b in self.bnds])):
            raise ValueError("Initial guess x0 exceeds upper bound constraints.")

        if include_unc is not None:
            self.include_unc = include_unc
        if unc_pmfs is not None:
            self.unc_pmfs = unc_pmfs
        if unc_options is not None:
            self.unc_options = unc_options

        if self.include_unc & (self.unc_pmfs is None):
            if self.unc_options is None:
                self.unc_options = {
                    "std_wd": 4.95,
                    "std_yaw": 1.75,
                    "pmf_res": 1.0,
                    "pdf_cutoff": 0.995,
                }

            # create normally distributed wd and yaw uncertaitny pmfs
            if self.unc_options["std_wd"] > 0:
                wd_bnd = int(
                    np.ceil(
                        norm.ppf(
                            self.unc_options["pdf_cutoff"],
                            scale=self.unc_options["std_wd"],
                        )
                        / self.unc_options["pmf_res"]
                    )
                )
                wd_unc = np.linspace(
                    -1 * wd_bnd * self.unc_options["pmf_res"],
                    wd_bnd * self.unc_options["pmf_res"],
                    2 * wd_bnd + 1,
                )
                wd_unc_pmf = norm.pdf(wd_unc, scale=self.unc_options["std_wd"])
                # normalize so sum = 1.0
                wd_unc_pmf = wd_unc_pmf / np.sum(wd_unc_pmf)
            else:
                wd_unc = np.zeros(1)
                wd_unc_pmf = np.ones(1)

            if self.unc_options["std_yaw"] > 0:
                yaw_bnd = int(
                    np.ceil(
                        norm.ppf(
                            self.unc_options["pdf_cutoff"],
                            scale=self.unc_options["std_yaw"],
                        )
                        / self.unc_options["pmf_res"]
                    )
                )
                yaw_unc = np.linspace(
                    -1 * yaw_bnd * self.unc_options["pmf_res"],
                    yaw_bnd * self.unc_options["pmf_res"],
                    2 * yaw_bnd + 1,
                )
                yaw_unc_pmf = norm.pdf(yaw_unc, scale=self.unc_options["std_yaw"])
                # normalize so sum = 1.0
                yaw_unc_pmf = yaw_unc_pmf / np.sum(yaw_unc_pmf)
            else:
                yaw_unc = np.zeros(1)
                yaw_unc_pmf = np.ones(1)

            self.unc_pmfs = {
                "wd_unc": wd_unc,
                "wd_unc_pmf": wd_unc_pmf,
                "yaw_unc": yaw_unc,
                "yaw_unc_pmf": yaw_unc_pmf,
            }

        if turbine_weights is None:
            self.turbine_weights = np.ones(self.nturbs)
        else:
            self.turbine_weights = np.array(turbine_weights, dtype=float)

        if calc_init_power:
            self._get_initial_farm_power()

        if exclude_downstream_turbines is not None:
            self.exclude_downstream_turbines = exclude_downstream_turbines
        self._reduce_control_variables()

    def calc_baseline_power(self):
        """
        This method computes the baseline power produced by the wind farm and
        the ideal power without wake losses for a series of wind speed, wind
        direction, and optionally TI combinations.

        Returns:
            pandas.DataFrame: A pandas DataFrame with the same number of rows
            as the length of the wd and ws arrays, containing the following
            columns:

                - **ws** (*float*) - The wind speed values for which power is
                computed (m/s).
                - **wd** (*float*) - The wind direction value for which power
                is calculated (deg).
                - **ti** (*float*) - The turbulence intensity value for which
                power is calculated. Only included if self.ti is not None.
                - **power_baseline** (*float*) - The total power produced by
                the wind farm with baseline yaw control (W).
                - **power_no_wake** (*float*) - The ideal total power produced
                by the wind farm without wake losses (W).
                - **turbine_power_baseline** (*list* (*float*)) - A
                list containing the baseline power without wake steering for
                each wind turbine in the wind farm (W).
                - **turbine_power_no_wake** (*list* (*float*)) - A list
                containing the ideal power without wake losses for each wind
                turbine in the wind farm (W).
        """
        print("=====================================================")
        print("Calculating baseline power...")
        print("Number of wind conditions to calculate = ", len(self.wd))
        print("=====================================================")

        # Put results in dict for speed, instead of previously
        # appending to frame.
        result_dict = dict()

        for i in range(len(self.wd)):
            if self.verbose:
                if self.ti is None:
                    print(
                        "Computing wind speed, wind direction pair "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg."
                    )
                else:
                    print(
                        "Computing wind speed, wind direction, turbulence "
                        + "intensity set "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg, turbulence intensity = "
                        + str(self.ti[i])
                        + "."
                    )

            # Find baseline power in FLORIS

            if self.ws[i] >= self.minimum_ws:
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )

                # calculate baseline power
                self.fi.calculate_wake(
                    yaw_angles=self.yaw_angles_baseline, no_wake=False
                )
                power_base = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )

                # calculate power for no wake case
                self.fi.calculate_wake(
                    yaw_angles=self.yaw_angles_baseline, no_wake=True
                )
                power_no_wake = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                    no_wake=True,
                )
            else:
                power_base = self.nturbs * [0.0]
                power_no_wake = self.nturbs * [0.0]

            # Include turbine weighing terms
            power_base = np.multiply(self.turbine_weights, power_base)
            power_no_wake = np.multiply(self.turbine_weights, power_no_wake)

            # add variables to dataframe
            if self.ti is None:
                result_dict[i] = {
                    "ws": self.ws[i],
                    "wd": self.wd[i],
                    "power_baseline": np.sum(power_base),
                    "turbine_power_baseline": power_base,
                    "power_no_wake": np.sum(power_no_wake),
                    "turbine_power_no_wake": power_no_wake,
                }
                # df_base = df_base.append(pd.DataFrame(
                #    {'ws':[self.ws[i]],'wd':[self.wd[i]],
                #    'power_baseline':[np.sum(power_base)],
                #    'turbine_power_baseline':[power_base],
                #    'power_no_wake':[np.sum(power_no_wake)],
                #    'turbine_power_no_wake':[power_no_wake]}))
            else:
                result_dict[i] = {
                    "ws": self.ws[i],
                    "wd": self.wd[i],
                    "ti": self.ti[i],
                    "power_baseline": np.sum(power_base),
                    "turbine_power_baseline": power_base,
                    "power_no_wake": np.sum(power_no_wake),
                    "turbine_power_no_wake": power_no_wake,
                }
                # df_base = df_base.append(pd.DataFrame(
                #    {'ws':[self.ws[i]],'wd':[self.wd[i]],
                #    'ti':[self.ti[i]],'power_baseline':[np.sum(power_base)],
                #    'turbine_power_baseline':[power_base],
                #    'power_no_wake':[np.sum(power_no_wake)],
                #    'turbine_power_no_wake':[power_no_wake]}))
        df_base = pd.DataFrame.from_dict(result_dict, "index")
        df_base.reset_index(drop=True, inplace=True)

        return df_base

    def optimize(self):
        """
        This method solves for the optimum turbine yaw angles for power
        production and the resulting power produced by the wind farm for a
        series of wind speed, wind direction, and optionally TI combinations.

        Returns:
            pandas.DataFrame: A pandas DataFrame with the same number of rows
            as the length of the wd and ws arrays, containing the following
            columns:

                - **ws** (*float*) - The wind speed values for which the yaw
                angles are optimized and power is computed (m/s).
                - **wd** (*float*) - The wind direction values for which the
                yaw angles are optimized and power is computed (deg).
                - **ti** (*float*) - The turbulence intensity values for which
                the yaw angles are optimized and power is computed. Only
                included if self.ti is not None.
                - **power_opt** (*float*) - The total power produced by the
                wind farm with optimal yaw offsets (W).
                - **turbine_power_opt** (*list* (*float*)) - A list
                containing the power produced by each wind turbine with optimal
                yaw offsets (W).
                - **yaw_angles** (*list* (*float*)) - A list containing
                the optimal yaw offsets for maximizing total wind farm power
                for each wind turbine (deg).
        """
        print("=====================================================")
        print("Optimizing wake redirection control...")
        print("Number of wind conditions to optimize = ", len(self.wd))
        print("Number of yaw angles to optimize = ", len(self.turbs_to_opt))
        print("=====================================================")

        df_opt = pd.DataFrame()

        for i in range(len(self.wd)):
            if self.verbose:
                if self.ti is None:
                    print(
                        "Computing wind speed, wind direction pair "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg."
                    )
                else:
                    print(
                        "Computing wind speed, wind direction, turbulence "
                        + "intensity set "
                        + str(i)
                        + " out of "
                        + str(len(self.wd))
                        + ": wind speed = "
                        + str(self.ws[i])
                        + " m/s, wind direction = "
                        + str(self.wd[i])
                        + " deg, turbulence intensity = "
                        + str(self.ti[i])
                        + "."
                    )

            # Optimizing wake redirection control
            if (self.ws[i] >= self.minimum_ws) & (self.ws[i] <= self.maximum_ws):
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )

                self.initial_farm_power = self.initial_farm_powers[i]
                opt_yaw_angles = self._optimize()

                if np.sum(opt_yaw_angles) == 0:
                    print(
                        "No change in controls suggested for this inflow \
                        condition..."
                    )

                # optimized power
                self.fi.calculate_wake(yaw_angles=opt_yaw_angles)
                power_opt = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            elif self.ws[i] >= self.maximum_ws:
                print(
                    "No change in controls suggested for this inflow \
                        condition..."
                )
                if self.ti is None:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
                    )
                else:
                    self.fi.reinitialize_flow_field(
                        wind_direction=[self.wd[i]],
                        wind_speed=[self.ws[i]],
                        turbulence_intensity=self.ti[i],
                    )
                opt_yaw_angles = np.array(self.yaw_angles_template, copy=True)
                self.fi.calculate_wake(yaw_angles=opt_yaw_angles)
                power_opt = self.fi.get_turbine_power(
                    include_unc=self.include_unc,
                    unc_pmfs=self.unc_pmfs,
                    unc_options=self.unc_options,
                )
            else:
                print(
                    "No change in controls suggested for this inflow \
                        condition..."
                )
                opt_yaw_angles = np.array(self.yaw_angles_template, copy=True)
                power_opt = self.nturbs * [0.0]

            # Include turbine weighing terms
            power_opt = np.multiply(self.turbine_weights, power_opt)

            # add variables to dataframe
            if self.ti is None:
                df_opt = df_opt.append(
                    pd.DataFrame(
                        {
                            "ws": [self.ws[i]],
                            "wd": [self.wd[i]],
                            "power_opt": [np.sum(power_opt)],
                            "turbine_power_opt": [power_opt],
                            "yaw_angles": [opt_yaw_angles],
                        }
                    )
                )
            else:
                df_opt = df_opt.append(
                    pd.DataFrame(
                        {
                            "ws": [self.ws[i]],
                            "wd": [self.wd[i]],
                            "ti": [self.ti[i]],
                            "power_opt": [np.sum(power_opt)],
                            "turbine_power_opt": [power_opt],
                            "yaw_angles": [opt_yaw_angles],
                        }
                    )
                )

        df_opt.reset_index(drop=True, inplace=True)

        return df_opt



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import copy

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.interpolate import griddata


def nudge_outward(x):
    """
    Avoid numerical issue in grid data by slightly expanding input x.
    TODO: expand this description
    - What numerical issues?
    - Whats the scenario when I would need this?

    Args:
        x (np.array): Vector of values.

    Returns:
        np.array: Expanded vector.
    """
    nudge_val = 0.001
    min_x = np.min(x)
    max_x = np.max(x)
    x = np.where(x == min_x, min_x - nudge_val, x)
    x = np.where(x == max_x, max_x + nudge_val, x)
    return x


def get_plane_from_flow_data(flow_data, normal_vector="z", x3_value=100):
    """
    Get a plane of data, in form of DataFrame, from a :py:class:`~.FlowData`
    object. This is used to get planes from SOWFA results and FLORIS
    simulations with fixed grids, i.e. curl.

    Args:
        flow_data (np.array): 3D vector field of velocity data. #TODO: is this
            supposed to be a :py:class:`~.FlowData` object?
        normal_vector (string, optional): Vector normal to plane.
            Defaults to z.
        x3_value (float, optional): Value of normal vector to slice through.
            Defaults to 100.

    Returns:
        pandas.DataFrame: Extracted data.
    """
    order = "f"
    if normal_vector == "z":
        x1_array = flow_data.x.flatten(order=order)
        x2_array = flow_data.y.flatten(order=order)
        x3_array = flow_data.z.flatten(order=order)

    if normal_vector == "x":
        x3_array = flow_data.x.flatten(order=order)
        x1_array = flow_data.y.flatten(order=order)
        x2_array = flow_data.z.flatten(order=order)

    if normal_vector == "y":
        x3_array = flow_data.y.flatten(order=order)
        x1_array = flow_data.x.flatten(order=order)
        x2_array = flow_data.z.flatten(order=order)

    u = flow_data.u.flatten(order=order)
    v = flow_data.v.flatten(order=order)
    w = flow_data.w.flatten(order=order)

    search_values = np.array(sorted(np.unique(x3_array)))
    nearest_idx = (np.abs(search_values - x3_value)).argmin()
    nearest_value = search_values[nearest_idx]
    print("Nearest value to %.2f is %.2f" % (x3_value, nearest_value))

    # Select down the data
    x3_select_mask = x3_array == nearest_value

    # Store the un-interpolated input arrays at this slice
    x1 = x1_array[x3_select_mask]
    x2 = x2_array[x3_select_mask]
    x3 = np.ones_like(x1) * x3_value
    u = u[x3_select_mask]
    v = v[x3_select_mask]
    w = w[x3_select_mask]

    df = pd.DataFrame({"x1": x1, "x2": x2, "x3": x3, "u": u, "v": v, "w": w})
    return df


class CutPlane:
    """
    A CutPlane object represents a 2D slice through the flow of a
    FLORIS simulation, or other such as SOWFA result.
    """

    def __init__(self, df, x1_resolution, x2_resolution):
        """
        Initialize CutPlane object, storing the DataFrame and resolution.

        Args:
            df (pandas.DataFrame): Pandas DataFrame of data with
                columns x1, x2, u, v, w.
        """
        self.df = df
        self.resolution = (x1_resolution, x2_resolution)


# Modification functions
def set_origin(cut_plane, center_x1=0.0, center_x2=0.0):
    """
    Establish the origin of a CutPlane object.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data.
        center_x1 (float, optional): x1-coordinate of origin.
            Defaults to 0.0.
        center_x2 (float, optional): x2-coordinate of origin.
            Defaults to 0.0.

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Updated plane of data.
    """
    # Store the un-interpolated input arrays at this slice
    cut_plane.df.x1 = cut_plane.df.x1 - center_x1
    cut_plane.df.x2 = cut_plane.df.x2 - center_x2

    return cut_plane


def change_resolution(cut_plane, resolution=(100, 100)):
    """
    Modify default resolution of a CutPlane object.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data.
        resolution (tuple, optional): Desired resolution in x1 and x2.
            Defaults to (100, 100).

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Updated plane of data.
    """

    # Linearize the data
    x1_lin = np.linspace(min(cut_plane.df.x1), max(cut_plane.df.x1), resolution[0])
    x2_lin = np.linspace(min(cut_plane.df.x2), max(cut_plane.df.x2), resolution[1])
    # x3 = np.ones_like(x1) * cut_plane.df.x3[0]

    # Mesh the data
    x1_mesh, x2_mesh = np.meshgrid(x1_lin, x2_lin)
    x3_mesh = np.ones_like(x1_mesh) * cut_plane.df.x3[0]

    # Interpolate u,v,w
    u_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.u.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )
    v_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.v.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )

    w_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.w.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )

    # Assign back to df
    cut_plane.df = pd.DataFrame(
        {
            "x1": x1_mesh.flatten(),
            "x2": x2_mesh.flatten(),
            "x3": x3_mesh.flatten(),
            "u": u_mesh.flatten(),
            "v": v_mesh.flatten(),
            "w": w_mesh.flatten(),
        }
    )

    # Save the resolution
    cut_plane.resolution = resolution

    # Return the cutplane
    return cut_plane


def interpolate_onto_array(cut_plane_in, x1_array, x2_array):
    """
    Interpolate a CutPlane object onto specified coordinate arrays.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data.
        x1_array (np.array): Specified x1-coordinate.
        x2_array (np.array): Specified x2-coordinate.

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Updated plane of data.
    """
    cut_plane = copy.deepcopy(cut_plane_in)

    # Linearize the data
    x1_lin = x1_array
    x2_lin = x2_array

    # Save the new resolution
    cut_plane.resolution = (len(np.unique(x1_lin)), len(np.unique(x2_lin)))

    # Mesh the data
    x1_mesh, x2_mesh = np.meshgrid(x1_lin, x2_lin)
    x3_mesh = np.ones_like(x1_mesh) * cut_plane.df.x3.iloc[0]

    # Interpolate u,v,w
    u_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.u.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )
    v_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.v.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )

    w_mesh = griddata(
        np.column_stack(
            [nudge_outward(cut_plane.df.x1), nudge_outward(cut_plane.df.x2)]
        ),
        cut_plane.df.w.values,
        (x1_mesh.flatten(), x2_mesh.flatten()),
        method="cubic",
    )

    # Assign back to df
    cut_plane.df = pd.DataFrame(
        {
            "x1": x1_mesh.flatten(),
            "x2": x2_mesh.flatten(),
            "x3": x3_mesh.flatten(),
            "u": u_mesh.flatten(),
            "v": v_mesh.flatten(),
            "w": w_mesh.flatten(),
        }
    )

    # Return the cutplane
    return cut_plane


def rescale_axis(cut_plane, x1_factor=1.0, x2_factor=1.0):
    """
    Stretch or compress CutPlane coordinates.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data.
        x1_factor (float): Scaling factor for x1-coordinate.
        x2_factor (float): Scaling factor for x2-coordinate.

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Updated plane of data.
    """
    # Store the un-interpolated input arrays at this slice
    cut_plane.df.x1 = cut_plane.df.x1 / x1_factor
    cut_plane.df.x2 = cut_plane.df.x2 / x2_factor

    return cut_plane


def project_onto(cut_plane_a, cut_plane_b):
    """
    Project cut_plane_a onto the x1, x2 of cut_plane_b

    Args:
        cut_plane_a (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data to project from.
        cut_plane_b (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data to project onto.

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Cut_plane_a projected onto cut_plane_b's axis.
    """

    return interpolate_onto_array(
        cut_plane_a, cut_plane_b.df.x1.unique(), cut_plane_b.df.x2.unique()
    )


def subtract(cut_plane_a_in, cut_plane_b_in):
    """
    Subtract u,v,w terms of cut_plane_b from cut_plane_a

    Args:
        cut_plane_a_in (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data to subtract from.
        cut_plane_b_in (:py:class:`~.tools.cut_plane.CutPlane`):
            Plane of data to subtract b.

    Returns:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            Difference of cut_plane_a_in minus cut_plane_b_in.
    """

    # First make copies of original
    cut_plane_a = copy.deepcopy(cut_plane_a_in)
    cut_plane_b = copy.deepcopy(cut_plane_b_in)

    # Sort x1 and x2 and make the index
    cut_plane_a.df = cut_plane_a.df.set_index(["x1", "x2"])
    cut_plane_b.df = cut_plane_b.df.set_index(["x1", "x2"])

    # Do subtraction
    cut_plane_a.df = cut_plane_a.df.subtract(
        cut_plane_b.df
    ).reset_index()  # .sort_values(['x2','x1'])# .dropna()
    # cut_plane_a.df = cut_plane_a.df.sort_values(['x1','x2'])
    return cut_plane_a


def calculate_wind_speed(cross_plane, x1_loc, x2_loc, R):
    """
    Calculate effective wind speed within specified range of a point.

    Args:
        cross_plane (:py:class:`floris.tools.cut_plane.CrossPlane`):
            plane of data.
        x1_loc (float): x1-coordinate of point of interst.
        x2_loc (float): x2-coordinate of point of interst.
        R (float): radius from point of interst to consider

    Returns:
        (float): effective wind speed
    """

    # Temp local copy
    df = cross_plane.df.copy()

    # Make a distance column
    df["distance"] = np.sqrt((df.x1 - x1_loc) ** 2 + (df.x2 - x2_loc) ** 2)

    # Return the cube-mean wind speed
    return np.cbrt(np.mean(df[df.distance < R].u ** 3))


def wind_speed_profile(cross_plane, R, x2_loc, resolution=100, x1_locs=None):

    if x1_locs is None:
        x1_locs = np.linspace(
            min(cross_plane.df.x1), max(cross_plane.df.x1), resolution
        )
    v_array = np.array(
        [calculate_wind_speed(cross_plane, x1_loc, x2_loc, R) for x1_loc in x1_locs]
    )
    return x1_locs, v_array


def calculate_power(
    cross_plane, x1_loc, x2_loc, R, ws_array, cp_array, air_density=1.225
):
    """
    Calculate maximum power available in a given cross plane.

    Args:
        cross_plane (:py:class:`floris.tools.cut_plane.CrossPlane`):
            plane of data.
        x1_loc (float): x1-coordinate of point of interst.
        x2_loc (float): x2-coordinate of point of interst.
        R (float): Radius of wind turbine rotor.
        ws_array (np.array): reference wind speed for cp curve.
        cp_array (np.array): cp curve at reference wind speeds.
        air_density (float, optional): air density. Defaults to 1.225.

    Returns:
        float: Power!
    """
    # Compute the ws
    ws = calculate_wind_speed(cross_plane, x1_loc, x2_loc, R)

    # Compute the cp
    cp_value = np.interp(ws, ws_array, cp_array)

    # Return the power
    return 0.5 * air_density * (np.pi * R ** 2) * cp_value * ws ** 3


def get_power_profile(
    cross_plane,
    x2_loc,
    ws_array,
    cp_array,
    R,
    air_density=1.225,
    resolution=100,
    x1_locs=None,
):

    if x1_locs is None:
        x1_locs = np.linspace(
            min(cross_plane.df.x1), max(cross_plane.df.x1), resolution
        )
    p_array = np.array(
        [
            calculate_power(
                cross_plane,
                x1_loc,
                x2_loc,
                R,
                ws_array,
                cp_array,
                air_density=air_density,
            )
            for x1_loc in x1_locs
        ]
    )
    return x1_locs, p_array


# # Define horizontal subclass
# class HorPlane(_CutPlane):
#     """
#     Subclass of _CutPlane. Shortcut to extracting a horizontal plane.
#     """

#     def __init__(self, df):
#         """
#         Initialize horizontal CutPlane

#         Args:
#             flow_data (np.array): 3D vector field of velocity data
#             z_value (float): vertical position through which to slice
#         """
#         # Set up call super
#         super().__init__(df)


# # Define cross plane subclass
# class CrossPlane(_CutPlane):
#     """
#     Subclass of _CutPlane. Shortcut to extracting a cross-stream plane.
#     """

#     def __init__(self, df):
#         """
#         Initialize cross-stream CutPlane

#         Args:
#             flow_data (np.array): 3D vector field of velocity data
#             x_value (float): streamwise position through which to slice
#         """
#         # Set up call super
#         super().__init__(df)


# # Define cross plane subclass
# class VertPlane(_CutPlane):
#     """
#     Subclass of _CutPlane. Shortcut to extracting a streamwise-vertical plane.
#     """

#     def __init__(self, df):
#         """
#         Initialize streamwise-vertical  CutPlane

#         Args:
#             flow_data (np.array): 3D vector field of velocity data
#             y_value (float): spanwise position through which to slice
#         """
#         # Set up call super
#         super().__init__(df)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
from pathlib import Path

from setuptools import setup, find_packages


# Package meta-data.
NAME = "FLORIS"
DESCRIPTION = "A controls-oriented engineering wake model."
URL = "https://github.com/NREL/FLORIS"
EMAIL = "rafael.mudafort@nrel.gov"
AUTHOR = "NREL National Wind Technology Center"
REQUIRES_PYTHON = ">=3.7.0"

# What packages are required for this module to be executed?
REQUIRED = [
    # simulation
    "attrs",
    "pyyaml",
    "numexpr",
    "numpy>=1.20",
    "scipy>=1.1",

    # tools
    "matplotlib>=3",
    "pandas",
    "shapely",

    # utilities
    "coloredlogs>=10.0",
]

# What packages are optional?
EXTRAS = {
    "docs": {"readthedocs-sphinx-ext", "Sphinx", "sphinxcontrib-napoleon"},
    "develop": {"pytest", "coverage[toml]", "pre-commit", "black", "isort"},
}

ROOT = Path(__file__).parent
with open(ROOT / "src" / "floris" / "VERSION") as version_file:
    VERSION = version_file.read().strip()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="Apache-2.0",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ],
)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


"""
The :py:obj:`floris` package contains :py:obj:`floris.utilities` module
and the modules that make up the FLORIS software. The floris simulation
modules are used to complete a wake simulation for a given wind farm
and turbine configuration.

All modules and package can be imported with

    >>> import floris

The ``__init__.py`` file enables the import of all modules in this
package so any additional modules should be included there.

isort:skip_file
"""

# Provide full-path imports here for all modules
# that should be included in the simulation package.
# Since some of these depend on each other, the order
# that they are listed here does matter.
from .base import BaseClass, BaseModel
from .turbine import Turbine, Ct, power, axial_induction, average_velocity
from .farm import Farm
from .grid import Grid, TurbineGrid, FlowFieldGrid, FlowFieldPlanarGrid
from .flow_field import FlowField
from .wake import WakeModelManager
from .solver import sequential_solver, full_flow_sequential_solver, cc_solver, full_flow_cc_solver
from .floris import Floris


# initialize the logger
import floris.logging_manager

floris.logging_manager._setup_logger()



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import matplotlib.pyplot as plt


class Layout:
    def __init__(self, fi, boundaries, wdir=None, wspd=None, wfreq=None):
        self.fi = fi
        self.boundaries = boundaries

        self.xmin = np.min([tup[0] for tup in boundaries])
        self.xmax = np.max([tup[0] for tup in boundaries])
        self.ymin = np.min([tup[1] for tup in boundaries])
        self.ymax = np.max([tup[1] for tup in boundaries])
        self.x0 = self.fi.layout_x
        self.y0 = self.fi.layout_y

        self.min_dist = 2 * self.rotor_diameter

        if wdir is not None:
            self.wdir = wdir
        else:
            self.wdir = self.fi.floris.farm.flow_field.wind_direction
        if wspd is not None:
            self.wspd = wspd
        else:
            self.wspd = self.fi.floris.farm.flow_field.wind_speed
        if wfreq is not None:
            self.wfreq = wfreq
        else:
            self.wfreq = 1.0

    def __str__(self):
        return "layout"

    ###########################################################################
    # Required private optimization methods
    ###########################################################################

    def reinitialize(self):
        pass

    def obj_func(self, varDict):
        # Parse the variable dictionary
        self.parse_opt_vars(varDict)

        # Update turbine map with turbince locations
        self.fi.reinitialize_flow_field(layout_array=[self.x, self.y])

        # Compute the objective function
        funcs = {}
        funcs["obj"] = (
            -1 * self.fi.get_farm_AEP(self.wdir, self.wspd, self.wfreq) * 1e-9
        )

        # Compute constraints, if any are defined for the optimization
        funcs = self.compute_cons(funcs)

        fail = False
        return funcs, fail

    # Optionally, the user can supply the optimization with gradients
    # def _sens(self, varDict, funcs):
    #     funcsSens = {}
    #     fail = False
    #     return funcsSens, fail

    def parse_opt_vars(self, varDict):
        self.x = varDict["x"]
        self.y = varDict["y"]

    def parse_sol_vars(self, sol):
        self.x = list(sol.getDVs().values())[0]
        self.y = list(sol.getDVs().values())[1]

    def add_var_group(self, optProb):
        optProb.addVarGroup(
            "x", self.nturbs, type="c", lower=self.xmin, upper=self.xmax, value=self.x0
        )
        optProb.addVarGroup(
            "y", self.nturbs, type="c", lower=self.ymin, upper=self.ymax, value=self.y0
        )

        return optProb

    def add_con_group(self, optProb):
        optProb.addConGroup("boundary_con", self.nturbs, lower=0.0)
        optProb.addConGroup("spacing_con", self.nturbs, lower=self.min_dist)

        return optProb

    def compute_cons(self, funcs):
        funcs["boundary_con"] = self.distance_from_boundaries()
        funcs["spacing_con"] = self.space_constraint()

        return funcs

    ###########################################################################
    # User-defined methods
    ###########################################################################

    def space_constraint(self):
        dist = [
            np.min(
                [
                    np.sqrt((self.x[i] - self.x[j]) ** 2 + (self.y[i] - self.y[j]) ** 2)
                    for j in range(self.nturbs)
                    if i != j
                ]
            )
            for i in range(self.nturbs)
        ]

        return dist

    def distance_from_boundaries(self):
        dist_out = []

        for k in range(self.nturbs):
            dist = []
            in_poly = self.point_inside_polygon(self.x[k], self.y[k], self.boundaries)

            for i in range(len(self.boundaries)):
                self.boundaries = np.array(self.boundaries)
                p1 = self.boundaries[i]
                if i == len(self.boundaries) - 1:
                    p2 = self.boundaries[0]
                else:
                    p2 = self.boundaries[i + 1]

                px = p2[0] - p1[0]
                py = p2[1] - p1[1]
                norm = px * px + py * py

                u = (
                    (self.x[k] - self.boundaries[i][0]) * px
                    + (self.y[k] - self.boundaries[i][1]) * py
                ) / float(norm)

                if u <= 0:
                    xx = p1[0]
                    yy = p1[1]
                elif u >= 1:
                    xx = p2[0]
                    yy = p2[1]
                else:
                    xx = p1[0] + u * px
                    yy = p1[1] + u * py

                dx = self.x[k] - xx
                dy = self.y[k] - yy
                dist.append(np.sqrt(dx * dx + dy * dy))

            dist = np.array(dist)
            if in_poly:
                dist_out.append(np.min(dist))
            else:
                dist_out.append(-np.min(dist))

        dist_out = np.array(dist_out)

        return dist_out

    def point_inside_polygon(self, x, y, poly):
        n = len(poly)
        inside = False

        p1x, p1y = poly[0]
        for i in range(n + 1):
            p2x, p2y = poly[i % n]
            if y > min(p1y, p2y):
                if y <= max(p1y, p2y):
                    if x <= max(p1x, p2x):
                        if p1y != p2y:
                            xinters = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                        if p1x == p2x or x <= xinters:
                            inside = not inside
            p1x, p1y = p2x, p2y

        return inside

    def plot_layout_opt_results(self, sol):
        """
        Method to plot the old and new locations of the layout opitimization.
        """
        locsx = sol.getDVs()["x"]
        locsy = sol.getDVs()["y"]

        plt.figure(figsize=(9, 6))
        fontsize = 16
        plt.plot(self.x0, self.y0, "ob")
        plt.plot(locsx, locsy, "or")
        # plt.title('Layout Optimization Results', fontsize=fontsize)
        plt.xlabel("x (m)", fontsize=fontsize)
        plt.ylabel("y (m)", fontsize=fontsize)
        plt.axis("equal")
        plt.grid()
        plt.tick_params(which="both", labelsize=fontsize)
        plt.legend(
            ["Old locations", "New locations"],
            loc="lower center",
            bbox_to_anchor=(0.5, 1.01),
            ncol=2,
            fontsize=fontsize,
        )

        verts = self.boundaries
        for i in range(len(verts)):
            if i == len(verts) - 1:
                plt.plot([verts[i][0], verts[0][0]], [verts[i][1], verts[0][1]], "b")
            else:
                plt.plot(
                    [verts[i][0], verts[i + 1][0]], [verts[i][1], verts[i + 1][1]], "b"
                )

        plt.show()

    ###########################################################################
    # Properties
    ###########################################################################

    @property
    def nturbs(self):
        """
        This property returns the number of turbines in the FLORIS
        object.

        Returns:
            nturbs (int): The number of turbines in the FLORIS object.
        """
        self._nturbs = len(self.fi.floris.farm.turbines)
        return self._nturbs

    @property
    def rotor_diameter(self):
        return self.fi.floris.farm.turbine_map.turbines[0].rotor_diameter



# functions to couple floris with CCBlade and a controller

import os
import copy
import pickle
from os import path

import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate

import floris.tools as wfct

from ..logging_manager import LoggerBase


# Attempt CCBlade import and raise error if no success
try:
    from ccblade import CCAirfoil, CCBlade
except ImportError:
    err_msg = (
        "CCBlade was not found. See http://wisdem.github.io/CCBlade for "
        + "installation instructions."
    )
    logger = LoggerBase()
    logger.logger.error(err_msg, stack_info=True)
    raise ImportError(err_msg)


# Some useful constants
degRad = np.pi / 180.0
rpmRadSec = 2.0 * (np.pi) / 60.0
base_R = 63.0  # Actual NREL 5MW radius


# Function returns a scaled NREL 5MW rotor object from CC-Blade
def CCrotor(
    Rtip=base_R,
    Rhub=1.5,
    hubHt=90.0,
    shearExp=0.2,
    rho=1.225,
    mu=1.81206e-5,
    path_to_af="5MW_AFFiles",
):

    r = (Rtip / base_R) * np.array(
        [
            2.8667,
            5.6000,
            8.3333,
            11.7500,
            15.8500,
            19.9500,
            24.0500,
            28.1500,
            32.2500,
            36.3500,
            40.4500,
            44.5500,
            48.6500,
            52.7500,
            56.1667,
            58.9000,
            61.6333,
        ]
    )
    chord = (Rtip / base_R) * np.array(
        [
            3.542,
            3.854,
            4.167,
            4.557,
            4.652,
            4.458,
            4.249,
            4.007,
            3.748,
            3.502,
            3.256,
            3.010,
            2.764,
            2.518,
            2.313,
            2.086,
            1.419,
        ]
    )
    theta = np.array(
        [
            13.308,
            13.308,
            13.308,
            13.308,
            11.480,
            10.162,
            9.011,
            7.795,
            6.544,
            5.361,
            4.188,
            3.125,
            2.319,
            1.526,
            0.863,
            0.370,
            0.106,
        ]
    )
    B = 3  # number of blades

    # In this initial version, hard-code to be NREL 5MW
    afinit = CCAirfoil.initFromAerodynFile  # just for shorthand
    basepath = path_to_af

    # load all airfoils
    airfoil_types = [0] * 8
    airfoil_types[0] = afinit(path.join(basepath, "Cylinder1.dat"))
    airfoil_types[1] = afinit(path.join(basepath, "Cylinder2.dat"))
    airfoil_types[2] = afinit(path.join(basepath, "DU40_A17.dat"))
    airfoil_types[3] = afinit(path.join(basepath, "DU35_A17.dat"))
    airfoil_types[4] = afinit(path.join(basepath, "DU30_A17.dat"))
    airfoil_types[5] = afinit(path.join(basepath, "DU25_A17.dat"))
    airfoil_types[6] = afinit(path.join(basepath, "DU21_A17.dat"))
    airfoil_types[7] = afinit(path.join(basepath, "NACA64_A17.dat"))

    # place at appropriate radial stations
    af_idx = [0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7]

    af = [0] * len(r)
    for i in range(len(r)):
        af[i] = airfoil_types[af_idx[i]]

    tilt = -5.0
    precone = 2.5
    yaw = 0.0

    nSector = 8  # azimuthal discretization

    rotor = CCBlade(
        r,
        chord,
        theta,
        af,
        Rhub,
        Rtip,
        B,
        rho,
        mu,
        precone,
        tilt,
        yaw,
        shearExp,
        hubHt,
        nSector,
    )

    return rotor


# Return the demanded generator torque for a given gen speed
# This is based on the torque controller within SOWFA and using the control parameters within the SOWFA example
def trq_cont(turbine_dict, genSpeedF):
    """
    Compute the torque control at a given gen speed (based on SOWFA)
    """
    # print(genSpeedF,turbine_dict['Region2StartGenSpeed'])
    # Region 1.
    if genSpeedF < turbine_dict["CutInGenSpeed"]:
        # print('in region 1...')
        torqueGenCommanded = turbine_dict["CutInGenTorque"]
    # # Region 1-1/2.
    elif (genSpeedF >= turbine_dict["CutInGenSpeed"]) and (
        genSpeedF < turbine_dict["Region2StartGenSpeed"]
    ):
        # print('in region 1.5...')
        dGenSpeed = genSpeedF - turbine_dict["CutInGenSpeed"]
        Region2StartGenTorque = (
            turbine_dict["KGen"]
            * turbine_dict["Region2StartGenSpeed"]
            * turbine_dict["Region2StartGenSpeed"]
        )
        torqueSlope = (Region2StartGenTorque - turbine_dict["CutInGenTorque"]) / (
            turbine_dict["Region2StartGenSpeed"] - turbine_dict["CutInGenSpeed"]
        )
        torqueGenCommanded = turbine_dict["CutInGenTorque"] + torqueSlope * dGenSpeed
    # # Region 2.
    elif (genSpeedF >= turbine_dict["Region2StartGenSpeed"]) and (
        genSpeedF < turbine_dict["Region2EndGenSpeed"]
    ):
        # print('in region 2...')
        torqueGenCommanded = turbine_dict["KGen"] * genSpeedF * genSpeedF
    # # Region 2-1/2.
    elif (genSpeedF >= turbine_dict["Region2EndGenSpeed"]) and (
        genSpeedF < turbine_dict["RatedGenSpeed"]
    ):
        # print('in region 2.5...')
        dGenSpeed = genSpeedF - turbine_dict["Region2EndGenSpeed"]
        Region2EndGenTorque = (
            turbine_dict["KGen"]
            * turbine_dict["Region2EndGenSpeed"]
            * turbine_dict["Region2EndGenSpeed"]
        )
        torqueSlope = (turbine_dict["RatedGenTorque"] - Region2EndGenTorque) / (
            turbine_dict["RatedGenSpeed"] - turbine_dict["Region2EndGenSpeed"]
        )
        torqueGenCommanded = Region2EndGenTorque + torqueSlope * dGenSpeed
    # # Region 3.
    elif genSpeedF >= turbine_dict["RatedGenSpeed"]:
        # print('in region 3...')
        torqueGenCommanded = turbine_dict["RatedGenTorque"]

    # Limit to the rated torque
    torqueGenCommanded = np.min([torqueGenCommanded, turbine_dict["RatedGenTorque"]])

    return torqueGenCommanded


# Update the PI pitch controller
# This is based on the pitch controller within SOWFA and using the control parameters within the SOWFA example
def pitch_control(turbine_dict, rotSpeedF, pitch_prev, dt, intSpeedError):
    min_pitch = 0.0
    max_pitch = 90.0

    # Set the gain scheduling variable.
    GK = 1.0 / (1.0 + (pitch_prev * degRad) / turbine_dict["PitchK"])

    # Store the old value of speed error.
    # speedErrorLast = sped_prev

    # Compute the low speed shaft speed error.
    speedError = rotSpeedF - turbine_dict["RatedRotSpeed"] * rpmRadSec  # in rad/s

    # Numerically integrate the speed error over time.
    intSpeedError = intSpeedError + speedError * dt

    # Numerically take the deriviative of speed error w.r.t time.
    # scalar derivSpeedError = (speedError[i] - speedErrorLast) / dt;

    # Saturate the integrated speed error based on pitch saturation.
    intSpeedError = np.max(
        [intSpeedError, min_pitch / (GK * turbine_dict["PitchControlKI"])]
    )
    intSpeedError = np.min(
        [intSpeedError, max_pitch / (GK * turbine_dict["PitchControlKI"])]
    )

    # Compute the pitch components from the proportional, integral,
    # and derivative parts and sum them.
    pitchP = GK * turbine_dict["PitchControlKP"] * speedError
    pitchI = GK * turbine_dict["PitchControlKI"] * intSpeedError
    # scalar pitchD = GK * PitchControlKD[j] * derivSpeedError;
    pitchCommanded = pitchP + pitchI  # + pitchD;

    # Saturate the pitch based on the pitch limits of the pitch
    # actuator.
    pitchCommanded = np.min([np.max([pitchCommanded, min_pitch]), max_pitch])

    # print('pitch commanded = ', pitchCommanded,max_pitch)

    # Return the commanded pitch
    return pitchCommanded, intSpeedError


# Given a controller paramaterized by turbing dict, return a new turbine_dict
# With values scaled according to changes in D and turbine rating (in MW)
def scale_controller_and_rotor(turbine_dict_in, R_In=base_R, turbine_rating=5):

    # Copy the dict
    turbine_dict = copy.deepcopy(turbine_dict_in)

    # Save the R value
    turbine_dict["TipRad"] = R_In

    # Scale the rotation speed inverse to the new radius
    turbine_dict["CutInGenSpeed"] = (base_R / R_In) * turbine_dict["CutInGenSpeed"]
    turbine_dict["Region2StartGenSpeed"] = (base_R / R_In) * turbine_dict[
        "Region2StartGenSpeed"
    ]
    turbine_dict["Region2EndGenSpeed"] = (base_R / R_In) * turbine_dict[
        "Region2EndGenSpeed"
    ]
    turbine_dict["RatedGenSpeed"] = (base_R / R_In) * turbine_dict["RatedGenSpeed"]
    turbine_dict["RatedRotSpeed"] = (base_R / R_In) * turbine_dict["RatedRotSpeed"]

    # Scale the cut in generator torque (not necessary, this is always 0)
    # turbine_dict['CutInGenTorque'] = (base_R/R_In) * turbine_dict['CutInGenTorque']

    # Scale kGen by the 5th power of radius
    turbine_dict["KGen"] = (R_In / base_R) ** 5 * turbine_dict["KGen"]

    # Scale the rator torque according to the rated speed and power
    turbine_dict["RatedGenTorque"] = (turbine_rating * 1e6) / (
        turbine_dict["RatedRotSpeed"]
        * turbine_dict["GBRatio"]
        * np.pi
        / 30.0
        * turbine_dict["GenEfficiency"]
    )

    # Save rating for conviencce
    turbine_dict["RatedMW"] = turbine_rating

    # Get the scaled rotor
    rotor = CCrotor(R_In)

    return turbine_dict, rotor


# Given a controller paramaterization, show the torque curve
def show_torque_curve(turbine_dict, ax, label="_nolegend_"):

    # Based on the details in SOWFA case, show the torque curve
    gen_speed_sweep = np.arange(
        0, turbine_dict["RatedRotSpeed"] * turbine_dict["GBRatio"], 1.0
    )
    gen_torque = np.array([trq_cont(turbine_dict, gf) for gf in gen_speed_sweep])
    # trq_opt = np.array([gf*gf*turbine_dict['KGen'] for gf in gen_speed_sweep])

    # ax.plot(gen_speed_sweep,trq_opt,'k--',label='Optimal')
    ax.plot(gen_speed_sweep, gen_torque, label=label)
    ax.set_xlabel("Gen Speed (RPM)")
    ax.set_ylabel("Gen Torque (Nm)")
    ax.grid(True)
    ax.set_title("Torque Curve")
    ax.legend()


# Generate a set of look-up tables the controller/steady state can use to find a cp/cq/ct
# for a given pitch angle and TSR
def generate_base_lut(rotor, turbine_dict):

    # These dicts (keyed on yaw)
    cp_dict = dict()
    ct_dict = dict()
    cq_dict = dict()

    # for now, assume only one yaw angle, perhaps expand later
    yaw = 0.0

    # Mesh the grid and flatten the arrays
    fixed_rpm = 10  # RPM
    Rtip = turbine_dict["TipRad"]
    TSR_initial = np.arange(0.5, 15, 0.5)
    pitch_initial = np.arange(0, 25, 0.5)
    ws_array = (fixed_rpm * (np.pi / 30.0) * Rtip) / TSR_initial
    ws_mesh, pitch_mesh = np.meshgrid(ws_array, pitch_initial)
    ws_flat = ws_mesh.flatten()
    pitch_flat = pitch_mesh.flatten()
    omega_flat = np.ones_like(pitch_flat) * fixed_rpm
    # tsr_flat = (fixed_rpm * (np.pi / 30.0) * Rtip) / ws_flat

    # Get values from cc-blade
    outputs, derivs = rotor.evaluate(
        [ws_flat], omega_flat, pitch_flat, coefficients=True
    )
    CP = outputs["CP"]
    CT = outputs["CT"]
    CQ = outputs["CQ"]

    # Reshape Cp, Ct and Cq
    CP = np.reshape(CP, (len(pitch_initial), len(TSR_initial)))
    CT = np.reshape(CT, (len(pitch_initial), len(TSR_initial)))
    CQ = np.reshape(CQ, (len(pitch_initial), len(TSR_initial)))

    # # Form the interpolant functions
    cp_interp = interpolate.interp2d(TSR_initial, pitch_initial, CP, kind="cubic")
    ct_interp = interpolate.interp2d(TSR_initial, pitch_initial, CT, kind="cubic")
    cq_interp = interpolate.interp2d(TSR_initial, pitch_initial, CQ, kind="cubic")

    # Add to the dictionaries
    cp_dict[yaw] = cp_interp
    ct_dict[yaw] = ct_interp
    cq_dict[yaw] = cq_interp

    # Save dictionaries
    pickle.dump([cp_dict, ct_dict, cq_dict], open("cp_ct_cq_lut.p", "wb"))


def get_aero_torque(rotor, ws, rot_speed, fluidDensity, R, pitch_angle=0.0):
    outputs, _ = rotor.evaluate(
        [ws], [rot_speed / rpmRadSec], [pitch_angle], coefficients=True
    )

    cq = outputs["CQ"]

    # print(cq[0])
    return 0.5 * fluidDensity * (np.pi * R ** 2) * cq[0] * R * ws ** 2


# For a given rotor/controller/wind speed, get the steady state value
def get_steady_state(
    turbine_dict, rotor, ws, dt=0.5, sim_time=5, title=None, show_plot=False
):

    # Save some convience terms
    fluidDensity = 1.225  # TODO Get from SOWFA
    R = turbine_dict["TipRad"]
    GBRatio = turbine_dict["GBRatio"]

    # Determine the drivetrain inertia
    drivetrain_inertia = (
        turbine_dict["NumBl"] * turbine_dict["BladeIner"]
        + turbine_dict["HubIner"]
        + turbine_dict["GBRatio"] * turbine_dict["GBRatio"] * turbine_dict["GenIner"]
    )

    # Simulation parameters
    # sim_length = sim_time/dt

    # Try to determine a good initial rotor speed
    rot_sweep = np.linspace(
        turbine_dict["CutInGenSpeed"] * rpmRadSec / GBRatio,
        turbine_dict["RatedRotSpeed"] * rpmRadSec,
        15,
    )
    gen_sweep = rot_sweep * GBRatio / rpmRadSec
    aero_sweep = np.array(
        [get_aero_torque(rotor, ws, r_speed, fluidDensity, R) for r_speed in rot_sweep]
    )
    gt_sweep = np.array([trq_cont(turbine_dict, gs) for gs in gen_sweep])
    torque_error = np.abs(
        aero_sweep * turbine_dict["GBEfficiency"] - GBRatio * gt_sweep
    )

    # If max exceeded, use max
    if np.max(aero_sweep * turbine_dict["GBEfficiency"]) > np.max(gt_sweep * GBRatio):
        init_rotor = turbine_dict["RatedRotSpeed"] * rpmRadSec
    else:  # Use the minimum
        idx = np.argmin(torque_error)
        init_rotor = rot_sweep[idx]

    # Initialize the pitch (if at max RPM)
    if (init_rotor == rot_sweep[-1]) or (
        init_rotor == turbine_dict["RatedRotSpeed"] * rpmRadSec
    ):
        pitch_sweep = np.linspace(0, 20, 50)
        aero_sweep = np.array(
            [
                get_aero_torque(rotor, ws, init_rotor, fluidDensity, R, pitch_angle=p)
                for p in pitch_sweep
            ]
        )
        gt_sweep = np.array(
            [trq_cont(turbine_dict, gen_sweep[-1]) for p in pitch_sweep]
        )
        torque_error = np.abs(
            aero_sweep * turbine_dict["GBEfficiency"] - GBRatio * gt_sweep
        )
        idx = np.argmin(torque_error)
        init_pitch = pitch_sweep[idx]

        # And force the intspeed error warm
        GK = 1.0 / (1.0 + (init_pitch * degRad) / turbine_dict["PitchK"])
        intSpeedError = init_pitch / (GK * turbine_dict["PitchControlKI"])

    else:
        init_pitch = 0.0
        # Initialize int speed error as 0
        intSpeedError = 0.0

    # Aero torque assuming pitch is 0

    # Create the arrays
    t_array = np.arange(0, sim_time, dt)
    pitch = np.ones_like(t_array) * init_pitch
    rot_speed = np.ones_like(t_array) * init_rotor  # represent rot speed in rad / s
    gen_speed = (
        np.ones_like(t_array) * init_rotor * GBRatio / rpmRadSec
    )  # represent gen speed in rpm
    aero_torque = np.ones_like(t_array) * 1000.0
    gen_torque = np.ones_like(t_array) * trq_cont(turbine_dict, gen_speed[0])
    gen_power = np.ones_like(t_array) * 0.0
    tsr_array = np.ones_like(t_array) * 0.0
    cq_array = np.ones_like(t_array) * 0.0
    cp_array = np.ones_like(t_array) * 0.0
    ct_array = np.ones_like(t_array) * 0.0

    # Load the Cp,Ct,Cq tables
    cp_dict, ct_dict, cq_dict = pickle.load(open("cp_ct_cq_lut.p", "rb"))

    # Select the 0-yaw LUT
    # cq_lut = cq_dict[0]

    # Now loop through and get the values
    re_run = True
    max_re_run = 5
    num_re_run = 0
    while re_run and (num_re_run < max_re_run):
        for i in range(1, len(t_array)):

            # print('Control time step = ', i, 'out of ', len(t_array))

            # Calculate TSR
            tsr = (R * (rot_speed[i - 1] / rpmRadSec) * np.pi / 30.0) / ws

            # Update the aero torque
            # cq = cq_lut(tsr,pitch[i-1])
            try:
                outputs, _ = rotor.evaluate(
                    [ws],
                    [rot_speed[i - 1] / rpmRadSec],
                    [pitch[i - 1]],
                    coefficients=True,
                )
                # M = outputs["M"]
                Cp = outputs["CP"]
                Ct = outputs["CT"]
                cq = outputs["CQ"]

            except Exception:
                print("CC BLADE PROBLEM")
                if i > 0:
                    return gen_power[i - 1], cp_array[i - 1], ct_array[i - 1]
                else:
                    print("...no data")
                    return np.nan, np.nan, np.nan

            aero_torque[i] = 0.5 * fluidDensity * (np.pi * R ** 2) * cq * R * ws ** 2

            # Save these values for plotting
            cq_array[i] = cq
            cp_array[i] = Cp[0]
            ct_array[i] = Ct[0]
            tsr_array[i] = tsr

            # Update the rotor speed and generator speed
            rot_speed[i] = rot_speed[i - 1] + (dt / drivetrain_inertia) * (
                aero_torque[i] * turbine_dict["GBEfficiency"]
                - GBRatio * gen_torque[i - 1]
            )
            gen_speed[i] = rot_speed[i] * GBRatio / rpmRadSec

            # Update the gen torque
            gen_torque[i] = trq_cont(turbine_dict, gen_speed[i])

            # Update the blade pitch
            pitch[i], intSpeedError = pitch_control(
                turbine_dict, rot_speed[i], pitch[i - 1], dt, intSpeedError
            )

            # Calculate the power
            gen_power[i] = (
                gen_speed[i]
                * np.pi
                / 30.0
                * gen_torque[i]
                * turbine_dict["GenEfficiency"]
            )
        # Determine if need to re_run
        if (gen_power[-1] <= turbine_dict["RatedMW"] * 1e6 * 1.001) and (
            np.abs(gen_torque[-1] - aero_torque[-1] / GBRatio) < 100
        ):
            re_run = False
        else:
            print("Re Run %s" % title)
            re_run = True
            num_re_run = num_re_run + 1
            pitch[0] = pitch[-1]
            rot_speed[0] = rot_speed[-1]
            gen_speed[0] = gen_speed[-1]
            aero_torque[0] = aero_torque[-1]
            gen_torque[0] = gen_torque[-1]
            gen_power[0] = gen_power[-1]
            tsr_array[0] = tsr_array[-1]
            cq_array[0] = cq_array[-1]

    if show_plot:
        fig, axarr = plt.subplots(5, 1, sharex=True)

        if title is not None:
            fig.suptitle(title, fontsize=16)

        ax = axarr[0]
        ax.plot(t_array[1:], tsr_array[1:], label="TSR")
        ax.fill_between(t_array[1:], 6, 8, color="gray", alpha=0.2)
        ax.legend()
        # ax.set_ylim([5,9])
        ax.grid(True)

        ax = axarr[1]
        ax.plot(t_array[1:], pitch[1:], label="Pitch")
        ax.grid(True)
        ax.legend()

        ax = axarr[2]
        ax.plot(t_array[1:], gen_torque[1:], label="Gen Torque")
        ax.plot(
            t_array[1:], aero_torque[1:] / GBRatio, label="Aero Torque / GB", color="r"
        )
        ax.grid(True)
        ax.legend()

        ax = axarr[3]
        ax.plot(t_array[1:], rot_speed[1:] / rpmRadSec, label="Rotor Speed (RPM)")
        ax.axhline(turbine_dict["RatedRotSpeed"], color="r", label="Rated")
        ax.grid(True)
        ax.legend()

        ax = axarr[4]
        ax.plot(t_array[1:], gen_power[1:] / 1e6, label="Power")
        ax.axhline(turbine_dict["RatedMW"], color="r")
        ax.plot()
        ax.grid(True)
        ax.legend()

    # Return the steady values
    return gen_power[-1], Cp[0], Ct[0]


def get_wind_sweep_steady_values(turbine_dict, rotor, ws_array=np.arange(3, 21, 1.0)):

    # Get the steady values
    pow_array = list()
    cp_array = list()
    ct_array = list()

    for ws in ws_array:
        print(ws)
        p, cp, ct = get_steady_state(turbine_dict, rotor, ws)
        pow_array.append(p)
        cp_array.append(cp)
        ct_array.append(ct)

    return ws_array, np.array(pow_array), np.array(cp_array), np.array(ct_array)



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


# Defines a bunch of tools for plotting and manipulating
# layouts for quick visualizations

import math

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from floris.utilities import wrap_360
from scipy.spatial.distance import pdist, squareform


# All functions assume a dataframe with index turbine, and columns x and y


def build_turbine_loc(turbine_x, turbine_y):
    """
    Make a DataFrame containing plant layout info
    (wind turbine locations).

    Args:
        turbine_x (np.array): wind turbine locations (east-west).
        turbine_y (np.array): wind turbine locations (north-south).

    Returns:
        turbineLoc (pd.DataFrame): turbine location data
    """
    turbineLoc = pd.DataFrame({"x": turbine_x, "y": turbine_y})
    return turbineLoc


def visualize_layout(
    turbineLoc,
    D,
    ax=None,
    show_wake_lines=False,
    limit_dist=None,
    turbine_face_north=False,
    one_index_turbine=False,
    black_and_white=False,
):
    """
    Make a plot which shows the turbine locations, and important wakes.

    Args:
        turbineLoc (pd.DataFrame): turbine location data
        D (float): wind turbine rotor diameter.
        ax (:py:class:`matplotlib.pyplot.axes` optional):
            figure axes. Defaults to None.
        show_wake_lines (bool, optional): flag to control plotting of
            wake boundaries. Defaults to False.
        limit_dist (float, optional): Downstream limit to plot wakes. D
            Defaults to None.
        turbine_face_north (bool, optional): Force orientation of wind
            turbines. Defaults to False.
        one_index_turbine (bool, optional): if true, 1st turbine is turbine 1
    """

    turbines = turbineLoc.index.values

    # if no axes provided, make one
    if not ax:
        fig, ax = plt.subplots(figsize=(7, 7))

    # Plot turbine points
    # ax.scatter(turbineLoc.x,turbineLoc.y,alpha=0)

    # Make ordered list of pairs sorted by distance if the distance
    # and angle matrices are provided
    if show_wake_lines:

        # Make a dataframe of distances
        dist = pd.DataFrame(
            squareform(pdist(turbineLoc)),
            index=turbineLoc.index,
            columns=turbineLoc.index,
        )

        # Make a DF of turbine angles
        angle = pd.DataFrame()
        turbines = turbineLoc.index
        for t1 in turbines:
            for t2 in turbines:
                # d[t1,t2] = wakeAngle(turbineLoc,[t1,t2])
                angle.loc[t1, t2] = wakeAngle(turbineLoc, [t1, t2])
        angle.index.name = "Turbine"

        # Now limit the matrix to only show waking from (row) to (column)
        for t1 in turbines:
            for t2 in turbines:
                if dist.loc[t1, t2] == 0.0:
                    dist.loc[t1, t2] = np.nan
                    angle.loc[t1, t2] = np.nan

        ordList = pd.DataFrame()
        for t1 in turbines:
            for t2 in turbines:
                temp = pd.DataFrame(
                    {
                        "T1": [t1],
                        "T2": [t2],
                        "Dist": [dist.loc[t1, t2]],
                        "angle": angle.loc[t1, t2],
                    }
                )
                ordList = pd.concat([ordList, temp])

        ordList.dropna(how="any", inplace=True)
        ordList.sort_values("Dist", inplace=True, ascending=False)

        # Plot wake lines and details
        for t1, t2 in zip(ordList.T1, ordList.T2):
            x = [turbineLoc.loc[t1, "x"], turbineLoc.loc[t2, "x"]]
            y = [turbineLoc.loc[t1, "y"], turbineLoc.loc[t2, "y"]]

            if limit_dist:
                if dist.loc[t1, t2] > limit_dist:
                    continue

            # Only plot positive x way
            if x[1] > x[0]:
                continue

            if black_and_white:
                (l,) = ax.plot(x, y, color="k")
            else:
                (l,) = ax.plot(x, y)
            # linetext = '%.2f m --- %.2f D --- %.2f Deg --- %.2f Deg' % (
            #     dist.loc[t1,t2],
            #     dist.loc[t1,t2]/D,
            #     angle.loc[t1,t2],
            #     angle.loc[t2,t1]
            # )
            # linetext = '%.2f D --- %.2f Deg' % (dist.loc[t1, t2] / D,
            #                                     angle.loc[t2, t1])
            linetext = "%.2f D --- %.1f/%.1f" % (
                dist.loc[t1, t2] / D,
                np.min([angle.loc[t2, t1], angle.loc[t1, t2]]),
                np.max([angle.loc[t2, t1], angle.loc[t1, t2]]),
            )
            # wrap_360(angle.loc[t2, t1]-180.))
            label_line(
                l, linetext, ax, near_i=1, near_x=None, near_y=None, rotation_offset=180
            )

    # Plot turbines
    for t1 in turbines:
        # print(t1)
        # ax.annotate(
        #     t1,
        #     (turbineLoc03.loc[t1].x,turbineLoc03.loc[t1].y),
        #     xycoords='data'
        # )
        if not turbine_face_north:
            ax.plot(
                [turbineLoc.loc[t1].x, turbineLoc.loc[t1].x],
                [
                    turbineLoc.loc[t1].y - 0.5 * D / 2.0,
                    turbineLoc.loc[t1].y + 0.5 * D / 2.0,
                ],
                color="k",
            )
        else:
            ax.plot(
                [
                    turbineLoc.loc[t1].x - 0.5 * D / 2.0,
                    turbineLoc.loc[t1].x + 0.5 * D / 2.0,
                ],
                [turbineLoc.loc[t1].y, turbineLoc.loc[t1].y],
                color="k",
            )
        if not one_index_turbine:
            ax.text(
                turbineLoc.loc[t1].x + D / 2,
                turbineLoc.loc[t1].y,
                t1,
                bbox=dict(boxstyle="round", ec="red", fc="white"),
            )
        else:
            ax.text(
                turbineLoc.loc[t1].x + D / 2,
                turbineLoc.loc[t1].y,
                t1 + 1,
                bbox=dict(boxstyle="round", ec="red", fc="white"),
            )
    ax.set_aspect("equal")


# Set wind direction
def set_direction(turbineLoc, rotation_angle):
    """
    Rotate wind farm CCW by the given angle provided in degrees

    #TODO add center of rotation? Default = center of farm?

    Args:
        turbineLoc (pd.DataFrame): turbine location data
        rotation_angle (float): rotation angle in degrees

    Returns:
        df_return (pd.DataFrame): rotated farm layout.
    """
    theta = np.deg2rad(rotation_angle)
    R = np.matrix([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    xy = np.array([turbineLoc.x, turbineLoc.y])

    xy_rot = R * xy
    # return xy_rot
    # print(xy_rot)
    # print(xy_rot[0][0][0])
    df_return = turbineLoc.copy(deep=True)
    df_return["x"] = np.squeeze(np.asarray(xy_rot[0, :]))
    df_return["y"] = np.squeeze(np.asarray(xy_rot[1, :]))
    return df_return


def turbineDist(df, turbList):
    """
    Derive distance between any two turbines.

    Args:
        df (pd.DataFrame): DataFrame with layout data.
        turbList (list): list of 2 turbines for which spacing distance
            is of interest.

    Returns:
        float: distance between turbines.
    """
    x1 = df.loc[turbList[0], "x"]
    x2 = df.loc[turbList[1], "x"]
    y1 = df.loc[turbList[0], "y"]
    y2 = df.loc[turbList[1], "y"]

    dist = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    return dist


def wakeAngle(df, turbList):
    """
    Get angles between turbines in wake direction

    Args:
        df (pd.DataFrame): DataFrame with layout data.
        turbList (list): list of 2 turbines for which spacing distance
            is of interest.

    Returns:
        wakeAngle (float): angle between turbines relative to compass
    """
    x1 = df.loc[turbList[0], "x"]
    x2 = df.loc[turbList[1], "x"]
    y1 = df.loc[turbList[0], "y"]
    y2 = df.loc[turbList[1], "y"]
    wakeAngle = (
        np.arctan2(y2 - y1, x2 - x1) * 180.0 / np.pi
    )  # Angle in normal cartesian coordinates

    # Convert angle to compass angle
    wakeAngle = 270.0 - wakeAngle
    if wakeAngle < 0:
        wakeAngle = wakeAngle + 360.0
    if wakeAngle > 360:
        wakeAngle = wakeAngle - 360.0

    return wakeAngle


def label_line(
    line,
    label_text,
    ax,
    near_i=None,
    near_x=None,
    near_y=None,
    rotation_offset=0.0,
    offset=(0, 0),
):
    """
    [summary]

    Args:
        line (matplotlib.lines.Line2D): line to label.
        label_text (str): label to add to line.
        ax (:py:class:`matplotlib.pyplot.axes` optional): figure axes.
        near_i (int, optional): Catch line near index i.
            Defaults to None.
        near_x (float, optional): Catch line near coordinate x.
            Defaults to None.
        near_y (float, optional): Catch line near coordinate y.
            Defaults to None.
        rotation_offset (float, optional): label rotation in degrees.
            Defaults to 0.
        offset (tuple, optional): label offset from turbine location.
            Defaults to (0, 0).

    Raises:
        ValueError: ("Need one of near_i, near_x, near_y") raised if
            insufficient information is passed in.
    """

    def put_label(i):
        """
        Add a label to index.

        Args:
            i (int): index to label.
        """
        i = min(i, len(x) - 2)
        dx = sx[i + 1] - sx[i]
        dy = sy[i + 1] - sy[i]
        rotation = np.rad2deg(math.atan2(dy, dx)) + rotation_offset
        pos = [(x[i] + x[i + 1]) / 2.0 + offset[0], (y[i] + y[i + 1]) / 2 + offset[1]]
        plt.text(
            pos[0],
            pos[1],
            label_text,
            size=9,
            rotation=rotation,
            color=line.get_color(),
            ha="center",
            va="center",
            bbox=dict(ec="1", fc="1", alpha=0.8),
        )

    # extract line data
    x = line.get_xdata()
    y = line.get_ydata()

    # define screen spacing
    if ax.get_xscale() == "log":
        sx = np.log10(x)
    else:
        sx = x
    if ax.get_yscale() == "log":
        sy = np.log10(y)
    else:
        sy = y

    # find index
    if near_i is not None:
        i = near_i
        if i < 0:  # sanitize negative i
            i = len(x) + i
        put_label(i)
    elif near_x is not None:
        for i in range(len(x) - 2):
            if (x[i] < near_x and x[i + 1] >= near_x) or (
                x[i + 1] < near_x and x[i] >= near_x
            ):
                put_label(i)
    elif near_y is not None:
        for i in range(len(y) - 2):
            if (y[i] < near_y and y[i + 1] >= near_y) or (
                y[i + 1] < near_y and y[i] >= near_y
            ):
                put_label(i)
    else:
        raise ValueError("Need one of near_i, near_x, near_y")


def make_turbine_array(x, y, filename="turbineArrayProperties", turbine="NREL5MWRef"):
    """
    Function to output a turbine array file given x and y locations.

    Args:
        x (np.array): wind turbine locations (east-west).
        y (np.array): wind turbine locations (north-south).
        filename (str, optional): write-to file path for wind turbine
            array info. Defaults to 'turbineArrayProperties'.
        turbine (str, optional): name of turbine to use within file.
            Defaults to 'NREL5MWRef'.
    """

    # Open the file for writing
    with open(filename, "w") as f:

        # Write out the headerlines
        f.write(
            "/*--------------------------------*- C++ -*----------------------------------*\\\n"
        )
        f.write(
            "| =========                 |                                                 |\n"
        )
        f.write(
            "| \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |\n"
        )
        f.write(
            "|  \\    /   O peration     | Version:  1.6                                   |\n"
        )
        f.write(
            "|   \\  /    A nd           | Web:      http://www.OpenFOAM.org               |\n"
        )
        f.write(
            "|    \\/     M anipulation  |                                                 |\n"
        )
        f.write(
            "\\*---------------------------------------------------------------------------*/\n"
        )
        f.write("FoamFile\n")
        f.write("{\n")
        f.write("    version     2.0;\n")
        f.write("    format      ascii;\n")
        f.write("    class       dictionary;\n")
        f.write("    object      turbineProperties;\n")
        f.write("}\n")
        f.write(
            "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
        )
        f.write("\n")
        f.write("globalProperties\n")
        f.write("{\n")
        f.write('    outputControl       "timeStep";\n')
        f.write("    outputInterval       1;\n")
        f.write("}\n")

        for idx, (x_val, y_val) in enumerate(zip(x, y)):
            f.write("\n")
            f.write("turbine%d\n" % idx)
            f.write("{\n")
            f.write('    turbineType         "%s";\n' % turbine)
            f.write("    baseLocation        (%.1f %.1f 0.0);\n" % (x_val, y_val))
            f.write("    nRadial              64;\n")
            f.write("    azimuthMaxDis        2.0;\n")
            f.write("    nAvgSector           1;\n")
            f.write('    pointDistType       "uniform";\n')
            f.write('    pointInterpType     "linear";\n')
            f.write('    bladeUpdateType     "oldPosition";\n')
            f.write("    epsilon              20.0;\n")
            f.write("    forceScalar          1.0;\n")
            f.write("    inflowVelocityScalar 0.94;\n")
            f.write('    tipRootLossCorrType "Glauert";\n')
            f.write('    rotationDir         "cw";\n')
            f.write("    Azimuth              0.0;\n")
            f.write("    RotSpeed             13.0;\n")
            f.write("    TorqueGen            20000.0;\n")
            f.write("    Pitch                0.0;\n")
            f.write("    NacYaw               270.0;\n")
            f.write("    fluidDensity         1.225;\n")
            f.write("}\n")



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numpy as np

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine
from floris.utilities import cosd, sind, tand


@define
class CrespoHernandez(BaseModel):
    """
    CrespoHernandez is a wake-turbulence model that is used to compute
    additional variability introduced to the flow field by operation of a wind
    turbine. Implementation of the model follows the original formulation and
    limitations outlined in :cite:`cht-crespo1996turbulence`.

    Args:
        parameter_dictionary (dict): Model-specific parameters.
            Default values are used when a parameter is not included
            in `parameter_dictionary`. Possible key-value pairs include:

            -   **initial** (*float*): The initial ambient turbulence
                intensity, expressed as a decimal fraction.
            -   **constant** (*float*): The constant used to scale the
                wake-added turbulence intensity.
            -   **ai** (*float*): The axial induction factor exponent used
                in in the calculation of wake-added turbulence.
            -   **downstream** (*float*): The exponent applied to the
                distance downstream of an upstream turbine normalized by
                the rotor diameter used in the calculation of wake-added
                turbulence.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: cht-
    """
    initial: float = field(converter=float, default=0.1)
    constant: float = field(converter=float, default=0.9)
    ai: float = field(converter=float, default=0.8)
    downstream: float = field(converter=float, default=-0.32)
    model_string = "crespo_hernandez"

    def prepare_function(self) -> dict:
        pass

    def function(
        self,
        ambient_TI: float,
        x: np.ndarray,
        x_i: np.ndarray,
        rotor_diameter: float,
        axial_induction: np.ndarray,
    ) -> None:
        # Replace zeros and negatives with 1 to prevent nans/infs
        delta_x = np.array(x - x_i)

        # TODO: ensure that these fudge factors are needed for different rotations
        upstream_mask = np.array(delta_x <= 0.1)
        downstream_mask = np.array(delta_x > -0.1)

        #        Keep downstream components          Set upstream to 1.0
        delta_x = delta_x * downstream_mask + np.ones_like(delta_x) * np.array(upstream_mask)

        # turbulence intensity calculation based on Crespo et. al.
        ti = (
            self.constant
        * axial_induction ** self.ai
        * ambient_TI ** self.initial
        * ((delta_x) / rotor_diameter) ** self.downstream
        )
        # Mask the 1 values from above with zeros
        return ti * np.array(downstream_mask)


# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation
from __future__ import annotations

from typing import Union
from itertools import product

import numpy as np
import matplotlib as mpl
import matplotlib.colors as mplcolors
import matplotlib.pyplot as plt
from matplotlib import rcParams


def plot_turbines(ax, layout_x, layout_y, yaw_angles, rotor_diameters, color=None, wind_direction=270.0):
    """
    Plot wind plant layout from turbine locations.

    Args:
        ax (:py:class:`matplotlib.pyplot.axes`): Figure axes.
        layout_x (np.array): Wind turbine locations (east-west).
        layout_y (np.array): Wind turbine locations (north-south).
        yaw_angles (np.array): Yaw angles of each wind turbine.
        D (float): Wind turbine rotor diameter.
        color (str): Pyplot color option to plot the turbines.
        wind_direction (float): Wind direction (rotates farm)
    """

    # Correct for the wind direction
    yaw_angles = np.array(yaw_angles)  # - wind_direction - 270

    if color is None:
        color = "k"
    for x, y, yaw, d in zip(layout_x, layout_y, yaw_angles, rotor_diameters):
        R = d / 2.0
        x_0 = x + np.sin(np.deg2rad(yaw)) * R
        x_1 = x - np.sin(np.deg2rad(yaw)) * R
        y_0 = y - np.cos(np.deg2rad(yaw)) * R
        y_1 = y + np.cos(np.deg2rad(yaw)) * R
        ax.plot([x_0, x_1], [y_0, y_1], color=color)


def plot_turbines_with_fi(ax, fi, color=None):
    """
    Wrapper function to plot turbines which extracts the data
    from a FLORIS interface object

    Args:
        ax (:py:class:`matplotlib.pyplot.axes`): figure axes. Defaults
            to None.
        fi (:py:class:`floris.tools.flow_data.FlowData`):
                FlowData object.
        color (str, optional): Color to plot turbines
    """
    plot_turbines(
        ax,
        fi.layout_x,
        fi.layout_y,
        fi.get_yaw_angles()[0, 0],
        fi.floris.farm.rotor_diameter[0, 0],
        color=color,
        wind_direction=fi.floris.flow_field.wind_directions[0],
    )


def line_contour_cut_plane(cut_plane, ax=None, levels=None, colors=None, **kwargs):
    """
    Visualize a cut_plane as a line contour plot.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`):
            CutPlane Object.
        ax (:py:class:`matplotlib.pyplot.axes`): Figure axes. Defaults
            to None.
        levels (np.array, optional): Contour levels for plot.
            Defaults to None.
        colors (list, optional): Strings of color specification info.
            Defaults to None.
        **kwargs: Additional parameters to pass to `ax.contour`.
    """

    if not ax:
        fig, ax = plt.subplots()

    # Reshape UMesh internally
    x1_mesh = cut_plane.df.x1.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    x2_mesh = cut_plane.df.x2.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    u_mesh = cut_plane.df.u.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    Zm = np.ma.masked_where(np.isnan(u_mesh), u_mesh)
    rcParams["contour.negative_linestyle"] = "solid"

    # # Plot the cut-through
    ax.contour(x1_mesh, x2_mesh, Zm, levels=levels, colors=colors, **kwargs)

    # Make equal axis
    ax.set_aspect("equal")


def visualize_cut_plane(
    cut_plane,
    ax=None,
    vel_component='u',
    minSpeed=None,
    maxSpeed=None,
    cmap="coolwarm",
    levels=None,
    color_bar=False,
    title=""
):
    """
    Generate pseudocolor mesh plot of the cut_plane.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`): 2D
            plane through wind plant.
        ax (:py:class:`matplotlib.pyplot.axes`): Figure axes. Defaults
            to None.
        minSpeed (float, optional): Minimum value of wind speed for
            contours. Defaults to None.
        maxSpeed (float, optional): Maximum value of wind speed for
            contours. Defaults to None.
        cmap (str, optional): Colormap specifier. Defaults to
            'coolwarm'.

    Returns:
        im (:py:class:`matplotlib.plt.pcolormesh`): Image handle.
    """

    if not ax:
        fig, ax = plt.subplots()
    if vel_component=='u':
        vel_mesh = cut_plane.df.u.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
        if minSpeed is None:
            minSpeed = cut_plane.df.u.min()
        if maxSpeed is None:
            maxSpeed = cut_plane.df.u.max()
    elif vel_component=='v':
        vel_mesh = cut_plane.df.v.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
        if minSpeed is None:
            minSpeed = cut_plane.df.v.min()
        if maxSpeed is None:
            maxSpeed = cut_plane.df.v.max()
    elif vel_component=='w':
        vel_mesh = cut_plane.df.w.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
        if minSpeed is None:
            minSpeed = cut_plane.df.w.min()
        if maxSpeed is None:
            maxSpeed = cut_plane.df.w.max()

    # Reshape to 2d for plotting
    x1_mesh = cut_plane.df.x1.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    x2_mesh = cut_plane.df.x2.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    Zm = np.ma.masked_where(np.isnan(vel_mesh), vel_mesh)

    # Plot the cut-through
    im = ax.pcolormesh(x1_mesh, x2_mesh, Zm, cmap=cmap, vmin=minSpeed, vmax=maxSpeed, shading="nearest")

    # Add line contour
    line_contour_cut_plane(cut_plane, ax=ax, levels=levels, colors="w", linewidths=0.8, alpha=0.3)

    if color_bar:
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label('m/s')

    # Set the title
    ax.set_title(title)

    # Make equal axis
    ax.set_aspect("equal")

    return im


def visualize_quiver(cut_plane, ax=None, minSpeed=None, maxSpeed=None, downSamp=1, **kwargs):
    """
    Visualize the in-plane flows in a cut_plane using quiver.

    Args:
        cut_plane (:py:class:`~.tools.cut_plane.CutPlane`): 2D
            plane through wind plant.
        ax (:py:class:`matplotlib.pyplot.axes`): Figure axes. Defaults
            to None.
        minSpeed (float, optional): Minimum value of wind speed for
            contours. Defaults to None.
        maxSpeed (float, optional): Maximum value of wind speed for
            contours. Defaults to None.
        downSamp (int, optional): Down sample the number of quiver arrows
            from underlying grid.
        **kwargs: Additional parameters to pass to `ax.streamplot`.

    Returns:
        im (:py:class:`matplotlib.plt.pcolormesh`): Image handle.
    """
    if not ax:
        fig, ax = plt.subplots()

    # Reshape UMesh internally
    x1_mesh = cut_plane.df.x1.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    x2_mesh = cut_plane.df.x2.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    v_mesh = cut_plane.df.v.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])
    w_mesh = cut_plane.df.w.values.reshape(cut_plane.resolution[1], cut_plane.resolution[0])

    # plot the stream plot
    ax.streamplot(
        (x1_mesh[::downSamp, ::downSamp]),
        (x2_mesh[::downSamp, ::downSamp]),
        v_mesh[::downSamp, ::downSamp],
        w_mesh[::downSamp, ::downSamp],
        # scale=80.0,
        # alpha=0.75,
        # **kwargs
    )

    # ax.quiverkey(QV1, -.75, -0.4, 1, '1 m/s', coordinates='data')

    # Make equal axis
    # ax.set_aspect('equal')


def reverse_cut_plane_x_axis_in_plot(ax):
    """
    Shortcut method to reverse direction of x-axis.

    Args:
        ax (:py:class:`matplotlib.pyplot.axes`): Figure axes.
    """
    ax.invert_xaxis()


def plot_rotor_values(
    values: np.ndarray,
    wd_index: int,
    ws_index: int,
    n_rows: int,
    n_cols: int,
    t_range: range | None = None,
    cmap: str = "coolwarm",
    return_fig_objects: bool = False,
    save_path: Union[str, None] = None,
    show: bool = False
) -> Union[None, tuple[plt.figure, plt.axes]]:
    """Plots the gridded turbine rotor values. This is intended to be used for
    understanding the differences between two sets of values, so each subplot can be
    used for inspection of what values are differing, and under what conditions.

    Parameters:
        values (np.ndarray): The 5-dimensional array of values to plot. Should be:
            N wind directions x N wind speeds x N turbines X N rotor points X N rotor points.
        cmap (str): The matplotlib colormap to be used, default "coolwarm".
        return_fig_objects (bool): Indicator to return the primary figure objects for
            further editing, default False.
        save_path (str | None): Where to save the figure, if a value is provided.

    Returns:
        None | tuple[plt.figure, plt.axes, plt.axis, plt.colorbar]: If
        `return_fig_objects` is `False, then `None` is returned`, otherwise the primary
        figure objects are returned for custom editing.

    Example:
        from floris.tools.visualization import plot_rotor_values
        plot_rotor_values(floris.flow_field.u, wd_range=range(0,1), ws_range=range(0,1))
        plot_rotor_values(floris.flow_field.v, wd_range=range(0,1), ws_range=range(0,1))
        plot_rotor_values(floris.flow_field.w, wd_range=range(0,1), ws_range=range(0,1), show=True)
    """

    cmap = plt.cm.get_cmap(name=cmap)

    if t_range is None:
        t_range = range(values.shape[2])

    fig = plt.figure()
    axes = fig.subplots(n_rows, n_cols)

    indices = t_range
    titles = np.array([f"T{i}" for i in indices])

    for ax, t, i in zip(axes.flatten(), titles, indices):

        vmin = np.min(values[wd_index, ws_index])
        vmax = np.max(values[wd_index, ws_index])

        bounds = np.linspace(vmin, vmax, 31)
        norm = mplcolors.Normalize(vmin, vmax)

        ax.imshow(values[wd_index, ws_index, i].T, cmap=cmap, norm=norm, origin="lower")

        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(t)

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([0.83, 0.25, 0.03, 0.5])
    cb = fig.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=cmap), cax=cbar_ax)

    if save_path:
        plt.savefig(save_path, bbox_inches="tight")

    if return_fig_objects:
        return fig, axes, cbar_ax, cb

    if show:
        plt.show()



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from typing import Any, Iterable, Tuple, Union, Callable

import attrs
from attrs import define, Attribute
import numpy as np
import numpy.typing as npt


### Define general data types used throughout

floris_float_type = np.float64

NDArrayFloat = npt.NDArray[floris_float_type]
NDArrayInt = npt.NDArray[np.int_]
NDArrayFilter = Union[npt.NDArray[np.int_], npt.NDArray[np.bool_]]
NDArrayObject = npt.NDArray[np.object_]


### Custom callables for attrs objects and functions

def floris_array_converter(data: Iterable) -> np.ndarray:
    try:
        a = np.array(data, dtype=floris_float_type)
    except TypeError as e:
        raise TypeError(e.args[0] + f". Data given: {data}")
    return a

def attr_serializer(inst: type, field: Attribute, value: Any):
    if isinstance(value, np.ndarray):
        return value.tolist()
    return value

def attr_floris_filter(inst: Attribute, value: Any) -> bool:
    if inst.init is False:
        return False
    if value is None:
        return False
    if isinstance(value, np.ndarray):
        if value.size == 0:
            return False
    return True

def iter_validator(iter_type, item_types: Union[Any, Tuple[Any]]) -> Callable:
    """Helper function to generate iterable validators that will reduce the amount of
    boilerplate code.

    Parameters
    ----------
    iter_type : any iterable
        The type of iterable object that should be validated.
    item_types : Union[Any, Tuple[Any]]
        The type or types of acceptable item types.

    Returns
    -------
    Callable
        The attr.validators.deep_iterable iterable and instance validator.
    """
    validator = attrs.validators.deep_iterable(
        member_validator=attrs.validators.instance_of(item_types),
        iterable_validator=attrs.validators.instance_of(iter_type),
    )
    return validator

@define
class FromDictMixin:
    """
    A Mixin class to allow for kwargs overloading when a data class doesn't
    have a specific parameter definied. This allows passing of larger dictionaries
    to a data class without throwing an error.
    """

    @classmethod
    def from_dict(cls, data: dict):
        """Maps a data dictionary to an `attr`-defined class.

        TODO: Add an error to ensure that either none or all the parameters are passed in

        Args:
            data : dict
                The data dictionary to be mapped.
        Returns:
            cls
                The `attr`-defined class.
        """
        # Check for any inputs that aren't part of the class definition
        class_attr_names = [a.name for a in cls.__attrs_attrs__]
        extra_args = [d for d in data if d not in class_attr_names]
        if len(extra_args):
            raise AttributeError(f"The initialization for {cls.__name__} was given extraneous inputs: {extra_args}")

        kwargs = {a.name: data[a.name] for a in cls.__attrs_attrs__ if a.name in data and a.init}

        # Map the inputs must be provided: 1) must be initialized, 2) no default value defined
        required_inputs = [a.name for a in cls.__attrs_attrs__ if a.init and a.default is attrs.NOTHING]
        undefined = sorted(set(required_inputs) - set(kwargs))
        if undefined:
            raise AttributeError(f"The class defintion for {cls.__name__} is missing the following inputs: {undefined}")
        return cls(**kwargs)

    def as_dict(self) -> dict:
        """Creates a JSON and YAML friendly dictionary that can be save for future reloading.
        This dictionary will contain only `Python` types that can later be converted to their
        proper `Turbine` formats.

        Returns:
            dict: All key, vaue pais required for class recreation.
        """
        return attrs.asdict(self, filter=attr_floris_filter, value_serializer=attr_serializer)


# Avoids constant redefinition of the same attr.ib properties for model attributes

# from functools import partial, update_wrapper

# def is_default(instance, attribute, value):
#     if attribute.default != value:
#         raise ValueError(f"{attribute.name} should never be set manually.")

# model_attrib = partial(field, on_setattr=attrs.setters.frozen, validator=is_default)  # type: ignore
# update_wrapper(model_attrib, field)

# float_attrib = partial(
#     attr.ib,
#     converter=float,
#     on_setattr=(attr.setters.convert, attr.setters.validate),  # type: ignore
#     kw_only=True,
# )
# update_wrapper(float_attrib, attr.ib)

# bool_attrib = partial(
#     attr.ib,
#     converter=bool,
#     on_setattr=(attr.setters.convert, attr.setters.validate),  # type: ignore
#     kw_only=True,
# )
# update_wrapper(bool_attrib, attr.ib)

# int_attrib = partial(
#     attr.ib,
#     converter=int,
#     on_setattr=(attr.setters.convert, attr.setters.validate),  # type: ignore
#     kw_only=True,
# )
# update_wrapper(int_attrib, attr.ib)




# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from ....logging_manager import LoggerBase


class Optimization(LoggerBase):
    """
    Base optimization class.

    Args:
        fi (:py:class:`floris.tools.floris_utilities.FlorisInterface`):
            Interface from FLORIS to the tools package.

    Returns:
        Optimization: An instantiated Optimization object.
    """

    def __init__(self, model, solver=None):
        """
        Instantiate Optimization object and its parameters.
        """
        self.model = model
        self.solver_choices = [
            "SNOPT",
            "IPOPT",
            "SLSQP",
            "NLPQLP",
            "FSQP",
            "NSGA2",
            "PSQP",
            "ParOpt",
            "CONMIN",
            "ALPSO",
        ]

        if solver not in self.solver_choices:
            raise ValueError(
                "Solver must be one supported by pyOptSparse: "
                + str(self.solver_choices)
            )

        self.reinitialize(solver=solver)

    # Private methods

    def _reinitialize(self, solver=None, optOptions=None):
        try:
            import pyoptsparse
        except ImportError:
            err_msg = (
                "It appears you do not have pyOptSparse installed. "
                + "Please refer to https://pyoptsparse.readthedocs.io/ for "
                + "guidance on how to properly install the module."
            )
            self.logger.error(err_msg, stack_info=True)
            raise ImportError(err_msg)

        self.optProb = pyoptsparse.Optimization(self.model, self.objective_func)

        self.optProb = self.model.add_var_group(self.optProb)
        self.optProb = self.model.add_con_group(self.optProb)
        self.optProb.addObj("obj")

        if solver is not None:
            self.solver = solver
            print("Setting up optimization with user's choice of solver: ", self.solver)
        else:
            self.solver = "SLSQP"
            print("Setting up optimization with default solver: SLSQP.")
        if optOptions is not None:
            self.optOptions = optOptions
        else:
            self.optOptions = {"Major feasibility tolerance": 1e-1}

        exec("self.opt = pyoptsparse." + self.solver + "(optOptions=self.optOptions)")

    def _optimize(self):
        if hasattr(self.model, "_sens"):
            self.sol = self.opt(self.optProb, sens=self.model._sens)
        else:
            self.sol = self.opt(self.optProb, sens="FDR")

    # Public methods

    def reinitialize(self, solver=None):
        self._reinitialize(solver=solver)

    def optimize(self):
        self._optimize()

        return self.sol

    def objective_func(self, varDict):
        return self.model.obj_func(varDict)

    def sensitivity_func(self):
        pass

    # Properties
